#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import json
import re
import time
import uuid
from collections import namedtuple, OrderedDict

from .expressions import *
from .dynamic import DynamicCollectionExpr
from .arithmetic import Negate
from . import utils
from ..types import validate_data_type, string, DynamicSchema
from ..utils import FunctionWrapper, output
from ...models import TableSchema
from ...compat import six, lkeys, lvalues, reduce
from ...utils import str_to_kv


class SortedColumn(SequenceExpr):
    """
    Notice: we do not inherit from the Column
    """

    __slots__ = '_ascending',
    _args = '_input',

    @property
    def input(self):
        return self._input

    def accept(self, visitor):
        return visitor.visit_sort_column(self)


class SortedExpr(Expr):
    def _init(self, *args, **kwargs):
        super(SortedExpr, self)._init(*args, **kwargs)

        if isinstance(self._ascending, bool):
            self._ascending = tuple([self._ascending] * len(self._sorted_fields))
        if len(self._sorted_fields) != len(self._ascending):
            raise ValueError('Length of ascending must be 1 or as many as the length of fields')

        sorted_fields = list()
        for i, field in enumerate(self._sorted_fields):
            if isinstance(field, Negate) and isinstance(field._input, SequenceExpr):
                field = field._input
                ascending = list(self._ascending)
                ascending[i] = False
                self._ascending = tuple(ascending)

                attr_values = dict((attr, getattr(field, attr, None)) for attr in utils.get_attrs(field))
                attr_values['_ascending'] = False
                sorted_fields.append(SortedColumn(**attr_values))
            elif isinstance(field, SortedColumn):
                sorted_fields.append(field)
            elif isinstance(field, SequenceExpr):
                column = SortedColumn(_input=field, _name=field.name, _source_name=field.source_name,
                                      _data_type=field._data_type,
                                      _source_data_type=field._source_data_type,
                                      _ascending=self._ascending[i])
                sorted_fields.append(column)
            elif isinstance(field, Scalar):
                column = SortedColumn(_input=field, _name=field.name, _source_name=field.source_name,
                                      _data_type=field._value_type,
                                      _source_data_type=field._source_value_type,
                                      _ascending=self._ascending[i])
                sorted_fields.append(column)
            else:
                from .groupby import SequenceGroupBy

                if isinstance(field, SequenceGroupBy):
                    field = field.name

                assert isinstance(field, six.string_types)
                sorted_fields.append(
                    SortedColumn(self._input[field], _name=field,
                                 _data_type=self._input._schema[field].type, _ascending=self._ascending[i]))
        self._sorted_fields = sorted_fields


class SortedCollectionExpr(SortedExpr, CollectionExpr):
    __slots__ = '_ascending',
    _args = '_input', '_sorted_fields'
    node_name = 'SortBy'

    def iter_args(self):
        for it in zip(['collection', 'keys'], self.args):
            yield it

    @property
    def input(self):
        return self._input

    def rebuild(self):
        rebuilt = super(SortedCollectionExpr, self).rebuild()
        rebuilt._schema = self.input.schema
        return rebuilt

    def accept(self, visitor):
        visitor.visit_sort(self)


def sort_values(expr, by, ascending=True):
    """
    Sort the collection by values. `sort` is an alias name for `sort_values`

    :param expr: collection
    :param by: the sequence or sequences to sort
    :param ascending: Sort ascending vs. descending. Sepecify list for multiple sort orders.
                      If this is a list of bools, must match the length of the by
    :return: Sorted collection

    :Example:

    >>> df.sort_values(['name', 'id'])  # 1
    >>> df.sort(['name', 'id'], ascending=False)  # 2
    >>> df.sort(['name', 'id'], ascending=[False, True])  # 3
    >>> df.sort([-df.name, df.id])  # 4, equal to #3
    """
    if not isinstance(by, list):
        by = [by, ]
    by = [it(expr) if inspect.isfunction(it) else it for it in by]
    return SortedCollectionExpr(expr, _sorted_fields=by, _ascending=ascending,
                                _schema=expr._schema)


class DistinctCollectionExpr(CollectionExpr):
    __slots__ = '_all',
    _args = '_input', '_unique_fields'
    node_name = 'Distinct'

    def _init(self, *args, **kwargs):
        super(DistinctCollectionExpr, self)._init(*args, **kwargs)

        if self._unique_fields:
            self._unique_fields = list(self._input._get_field(field)
                                       for field in self._unique_fields)

            if not hasattr(self, '_schema'):
                names = [field.name for field in self._unique_fields]
                types = [field._data_type for field in self._unique_fields]
                self._schema = TableSchema.from_lists(names, types)
        else:
            self._unique_fields = list(self._input._get_field(field)
                                       for field in self._input._schema.names)
            self._schema = self._input._schema

    def iter_args(self):
        for it in zip(['collection', 'distinct'], self.args):
            yield it

    @property
    def input(self):
        return self._input

    def rebuild(self):
        return self._input.distinct(self._unique_fields if not self._all else [])

    def accept(self, visitor):
        return visitor.visit_distinct(self)


def distinct(expr, on=None, *ons):
    """
    Get collection with duplicate rows removed, optionally only considering certain columns

    :param expr: collection
    :param on: sequence or sequences
    :return: dinstinct collection

    :Example:

    >>> df.distinct(['name', 'id'])
    >>> df['name', 'id'].distinct()
    """

    on = on or list()
    if not isinstance(on, list):
        on = [on, ]
    on = on + list(ons)

    on = [it(expr) if inspect.isfunction(it) else it for it in on]

    return DistinctCollectionExpr(expr, _unique_fields=on, _all=(len(on) == 0))


def unique(expr):
    if isinstance(expr, SequenceExpr):
        collection = next(it for it in expr.traverse(top_down=True, unique=True)
                          if isinstance(it, CollectionExpr))
        return collection.distinct(expr)[expr.name]


class SampledCollectionExpr(CollectionExpr):
    _args = '_input', '_n', '_frac', '_parts', '_i', '_sampled_fields', '_replace', \
            '_weights', '_strata', '_random_state'
    node_name = 'Sample'

    def _init(self, *args, **kwargs):
        for attr in self._args[1:]:
            self._init_attr(attr, None)
        super(SampledCollectionExpr, self)._init(*args, **kwargs)

        if not isinstance(self._n, dict):
            self._n = self._scalar(self._n)
        else:
            self._n = self._scalar(json.dumps(self._n))
        if not isinstance(self._frac, dict):
            self._frac = self._scalar(self._frac)
        else:
            self._frac = self._scalar(json.dumps(self._frac))
        self._parts = self._scalar(self._parts)
        self._i = self._scalar(self._i)
        self._replace = self._scalar(self._replace)
        self._weights = self._scalar(self._weights)
        self._strata = self._scalar(self._strata)
        self._random_state = self._scalar(self._random_state)

    def _scalar(self, val):
        if val is None:
            return
        if isinstance(val, Scalar):
            return val
        if isinstance(val, tuple):
            return tuple(self._scalar(it) for it in val)
        else:
            return Scalar(_value=val)

    @property
    def input(self):
        return self._input

    def rebuild(self):
        rebuilt = super(SampledCollectionExpr, self).rebuild()
        rebuilt._schema = self.input.schema
        return rebuilt

    def accept(self, visitor):
        return visitor.visit_sample(self)


def __df_sample(expr, parts=None, columns=None, i=None, n=None, frac=None, replace=False,
                 weights=None, strata=None, random_state=None):
    if columns:
        columns = expr.select(columns)._fields

    return SampledCollectionExpr(_input=expr, _parts=parts, _i=i, _sampled_fields=columns, _n=n,
                                 _frac=frac, _weights=weights, _strata=strata, _random_state=random_state,
                                 _replace=replace, _schema=expr.schema)


def sample(expr, parts=None, columns=None, i=None, n=None, frac=None, replace=False,
           weights=None, strata=None, random_state=None):
    """
    Sample collection.

    :param expr: collection
    :param parts: how many parts to hash
    :param columns: the columns to sample
    :param i: the part to sample out, can be a list of parts, must be from 0 to parts-1
    :param n: how many rows to sample. If `strata` is specified, `n` should be a dict with values in the strata column as dictionary keys and corresponding sample size as values
    :param frac: how many fraction to sample. If `strata` is specified, `n` should be a dict with values in the strata column as dictionary keys and corresponding sample weight as values
    :param replace: whether to perform replace sampling
    :param weights: the column name of weights
    :param strata: the name of strata column
    :param random_state: the random seed when performing sampling
    :return: collection

    Note that n, frac, replace, weights, strata and random_state can only be used under Pandas DataFrames or
    XFlow.

    :Example:

    Sampling with parts:

    >>> df.sample(parts=1)
    >>> df.sample(parts=5, i=0)
    >>> df.sample(parts=10, columns=['name'])

    Sampling with fraction or weights, replacement option can be specified:

    >>> df.sample(n=100)
    >>> df.sample(frac=0.1)
    >>> df.sample(frac=0.1, replace=True)

    Sampling with weight column:

    >>> df.sample(n=100, weights='weight_col')
    >>> df.sample(n=100, weights='weight_col', replace=True)

    Stratified sampling. Note that currently we do not support stratified sampling with replacement.

    >>> df.sample(strata='category', frac={'Iris Setosa': 0.5, 'Iris Versicolour': 0.4})
    """
    if isinstance(expr, CollectionExpr):
        if n is None and frac is None and parts is None:
            raise ExpressionError('Either n or frac or parts should be provided')
        if i is not None and parts is None:
            raise ExpressionError('`parts` arg is required when `i` arg is specified')
        if len([arg for arg in (n, frac, parts) if arg is not None]) > 1:
            raise ExpressionError('You cannot specify `n` or `frac` or `parts` at the same time')
        if strata is None and n is not None and frac is not None:
            # strata can specify different types of strategies on different columns
            raise ExpressionError('You cannot specify `n` and `frac` at the same time.')
        if weights is not None and strata is not None:
            raise ExpressionError('You cannot specify `weights` and `strata` at the same time.')
        if strata is not None:
            if frac is not None and not isinstance(frac, (six.string_types, dict)):
                raise ExpressionError('`frac` should be a k-v string or a dictionary object.')
            if isinstance(frac, six.string_types):
                frac = str_to_kv(frac, float)

            if n is not None and not isinstance(n, (six.string_types, dict)):
                raise ExpressionError('`n` should be a k-v string or a dictionary object.')
            if isinstance(n, six.string_types):
                n = str_to_kv(n, int)

            for val in six.itervalues(frac or dict()):
                if val < 0 or val > 1:
                    raise ExpressionError('Values in `frac` must be between 0 and 1')
            if n is not None and frac is not None:
                collides = set(six.iterkeys(n)).intersection(set(six.iterkeys(frac)))
                if collides:
                    raise ExpressionError('Values in `frac` and `n` collides with each other.')
        else:
            if frac is not None and (not isinstance(frac, (six.integer_types, float)) or frac < 0 or frac > 1):
                raise ExpressionError('`frac` must be between 0 and 1')

        if parts is not None:
            if i is None:
                i = (0, )
            elif isinstance(i, list):
                i = tuple(i)
            elif not isinstance(i, tuple):
                i = (i, )

            for it in i:
                if it >= parts or it < 0:
                    raise ExpressionError('`i` should be positive numbers that less than `parts`')
        elif not options.df.use_xflow_sample and not replace and weights is None and strata is None:
            if frac is not None and frac < 0.01:
                raise ValueError(
                    "Does not support sampling less than 1%. Try sampling by count or "
                    "set options.df.use_xflow_sample to True."
                )
        elif hasattr(expr, '_xflow_sample'):
            return expr._xflow_sample(n=n, frac=frac, replace=replace, weights=weights, strata=strata,
                                      random_state=random_state)

        return expr.__sample(parts=parts, columns=columns, i=i, n=n, frac=frac, replace=replace,
                             weights=weights, strata=strata, random_state=random_state)


class RowAppliedCollectionExpr(CollectionExpr):
    __slots__ = '_func', '_func_args', '_func_kwargs', '_close_func', \
                '_resources', '_raw_inputs', '_lateral_view', '_keep_nulls'
    _args = '_input', '_fields', '_collection_resources'
    node_name = 'Apply'

    def _init(self, *args, **kwargs):
        self._init_attr('_raw_inputs', None)
        self._init_attr('_lateral_view', False)
        super(RowAppliedCollectionExpr, self)._init(*args, **kwargs)

    @property
    def input(self):
        return self._input

    @property
    def fields(self):
        return self._fields

    @property
    def input_types(self):
        return [f.dtype for f in self._fields]

    @property
    def raw_input_types(self):
        if self._raw_inputs:
            return [f.dtype for f in self._raw_inputs]
        return self.input_types

    @property
    def func(self):
        return self._func

    @func.setter
    def func(self, f):
        self._func = f

    def accept(self, visitor):
        return visitor.visit_apply_collection(self)


def _apply_horizontal(expr, func, names=None, types=None, resources=None,
                      collection_resources=None, keep_nulls=False,
                      args=(), **kwargs):
    if isinstance(func, FunctionWrapper):
        names = names or func.output_names
        types = types or func.output_types
        func = func._func

    if names is not None:
        if isinstance(names, list):
            names = tuple(names)
        elif isinstance(names, six.string_types):
            names = (names,)

    if names is None:
        raise ValueError(
            'Apply on rows to provide multiple values should provide all column names, '
            'for instance, df.apply(func, axis=1, names=["A", "B"], types=["float", "float"]). '
            'See https://pyodps.readthedocs.io/zh_CN/latest/df-sort-distinct-apply.html#dfudtfapp '
            'for more information.'
        )
    tps = (string,) * len(names) if types is None else tuple(validate_data_type(t) for t in types)
    schema = TableSchema.from_lists(names, tps)

    collection_resources = collection_resources or \
                           utils.get_collection_resources(resources)
    return RowAppliedCollectionExpr(_input=expr, _func=func, _func_args=args,
                                    _func_kwargs=kwargs, _schema=schema,
                                    _fields=[expr[n] for n in expr.schema.names],
                                    _keep_nulls=keep_nulls, _resources=resources,
                                    _collection_resources=collection_resources)


def apply(expr, func, axis=0, names=None, types=None, reduce=False,
          resources=None, keep_nulls=False, args=(), **kwargs):
    """
    Apply a function to a row when axis=1 or column when axis=0.

    :param expr:
    :param func: function to apply
    :param axis: row when axis=1 else column
    :param names: output names
    :param types: output types
    :param reduce: if True will return a sequence else return a collection
    :param resources: resources to read
    :param keep_nulls: if True, keep rows producing empty results, only work in lateral views
    :param args: args for function
    :param kwargs: kwargs for function
    :return:

    :Example:

    Apply a function to a row:

    >>> from odps.df import output
    >>>
    >>> @output(['iris_add', 'iris_sub'], ['float', 'float'])
    >>> def handle(row):
    >>>     yield row.sepallength - row.sepalwidth, row.sepallength + row.sepalwidth
    >>>     yield row.petallength - row.petalwidth, row.petallength + row.petalwidth
    >>>
    >>> iris.apply(handle, axis=1).count()


    Apply a function to a column:

    >>> class Agg(object):
    >>>
    >>>     def buffer(self):
    >>>         return [0.0, 0]
    >>>
    >>>     def __call__(self, buffer, val):
    >>>         buffer[0] += val
    >>>         buffer[1] += 1
    >>>
    >>>     def merge(self, buffer, pbuffer):
    >>>         buffer[0] += pbuffer[0]
    >>>         buffer[1] += pbuffer[1]
    >>>
    >>>     def getvalue(self, buffer):
    >>>         if buffer[1] == 0:
    >>>             return 0.0
    >>>         return buffer[0] / buffer[1]
    >>>
    >>> iris.exclude('name').apply(Agg)
    """

    if types is None and "rtype" in kwargs:
        types = kwargs.pop("rtype")

    if not isinstance(expr, CollectionExpr):
        return

    if isinstance(func, FunctionWrapper):
        names = names or func.output_names
        types = types or func.output_types
        func = func._func

    if axis == 0:
        types = types or expr.schema.types
        types = [validate_data_type(t) for t in types]

        fields = [expr[n].agg(func, rtype=t, resources=resources)
                  for n, t in zip(expr.schema.names, types)]
        if names:
            fields = [f.rename(n) for f, n in zip(fields, names)]
        else:
            names = [f.name for f in fields]
        return Summary(_input=expr, _fields=fields, _schema=TableSchema.from_lists(names, types))
    else:
        collection_resources = utils.get_collection_resources(resources)

        if types is not None:
            if isinstance(types, list):
                types = tuple(types)
            elif isinstance(types, six.string_types):
                types = (types,)

            types = tuple(validate_data_type(t) for t in types)
        if reduce:
            from .element import MappedExpr
            from ..backends.context import context

            if names is not None and len(names) > 1:
                raise ValueError('When reduce, at most one name can be specified')
            name = names[0] if names is not None else None
            if not types and kwargs.get('rtype', None) is not None:
                types = [kwargs.pop('rtype')]
            tp = types[0] if types is not None else (utils.get_annotation_rtype(func) or string)
            if not context.is_cached(expr) and (hasattr(expr, '_fields') and expr._fields is not None):
                inputs = [e.copy_tree(stop_cond=lambda x: any(i is expr.input for i in x.children()))
                          for e in expr._fields]
            else:
                inputs = [expr[n] for n in expr.schema.names]
            return MappedExpr(_func=func, _func_args=args, _func_kwargs=kwargs,
                              _name=name, _data_type=tp,
                              _inputs=inputs, _multiple=True,
                              _resources=resources, _collection_resources=collection_resources)
        else:
            return _apply_horizontal(expr, func, names=names, types=types, resources=resources,
                                     collection_resources=collection_resources, keep_nulls=keep_nulls,
                                     args=args, **kwargs)


class ReshuffledCollectionExpr(CollectionExpr):
    _args = '_input', '_by', '_sort_fields'
    node_name = 'Reshuffle'

    def _init(self, *args, **kwargs):
        from .groupby import BaseGroupBy, SortedGroupBy

        self._init_attr('_sort_fields', None)

        super(ReshuffledCollectionExpr, self)._init(*args, **kwargs)

        if isinstance(self._input, BaseGroupBy):
            if isinstance(self._input, SortedGroupBy):
                self._sort_fields = self._input._sorted_fields
            self._by = self._input._by
            self._input = self._input._input

    @property
    def fields(self):
        return self._by + (self._sort_fields or list())

    @property
    def input(self):
        return self._input

    def iter_args(self):
        arg_names = ['collection', 'bys', 'sort']
        for it in zip(arg_names, self.args):
            yield it

    def accept(self, visitor):
        return visitor.visit_reshuffle(self)


def reshuffle(expr, by=None, sort=None, ascending=True):
    """
    Reshuffle data.

    :param expr:
    :param by: the sequence or scalar to shuffle by. RandomScalar as default
    :param sort: the sequence or scalar to sort.
    :param ascending: True if ascending else False
    :return: collection
    """

    by = by or RandomScalar()

    grouped = expr.groupby(by)
    if sort:
        grouped = grouped.sort_values(sort, ascending=ascending)

    return ReshuffledCollectionExpr(_input=grouped, _schema=expr._schema)


def map_reduce(expr, mapper=None, reducer=None, group=None, sort=None, ascending=True,
               combiner=None, combiner_buffer_size=1024,
               mapper_output_names=None, mapper_output_types=None, mapper_resources=None,
               reducer_output_names=None, reducer_output_types=None, reducer_resources=None):
    """
    MapReduce API, mapper or reducer should be provided.

    :param expr:
    :param mapper: mapper function or class
    :param reducer: reducer function or class
    :param group: the keys to group after mapper
    :param sort: the keys to sort after mapper
    :param ascending: True if ascending else False
    :param combiner: combiner function or class, combiner's output should be equal to mapper
    :param combiner_buffer_size: combiner's buffer size, 1024 as default
    :param mapper_output_names: mapper's output names
    :param mapper_output_types: mapper's output types
    :param mapper_resources: the resources for mapper
    :param reducer_output_names: reducer's output names
    :param reducer_output_types: reducer's output types
    :param reducer_resources: the resources for reducer
    :return:

    :Example:

    >>> from odps.df import output
    >>>
    >>> @output(['word', 'cnt'], ['string', 'int'])
    >>> def mapper(row):
    >>>     for word in row[0].split():
    >>>         yield word.lower(), 1
    >>>
    >>> @output(['word', 'cnt'], ['string', 'int'])
    >>> def reducer(keys):
    >>>     cnt = [0]
    >>>     def h(row, done):  # done illustrates that all the rows of the keys are processed
    >>>         cnt[0] += row.cnt
    >>>         if done:
    >>>             yield keys.word, cnt[0]
    >>>     return h
    >>>
    >>> words_df.map_reduce(mapper, reducer, group='word')
    """

    def _adjust_partial(fun):
        if isinstance(fun, functools.partial) and isinstance(fun.func, FunctionWrapper):
            wrapped_fun = fun.func
            partial_fun = functools.partial(wrapped_fun._func, *fun.args, **fun.keywords)
            ret_fun = FunctionWrapper(partial_fun)
            ret_fun.output_names = wrapped_fun.output_names
            ret_fun.output_types = wrapped_fun.output_types
            return ret_fun
        else:
            return fun

    def conv(l):
        if l is None:
            return
        if isinstance(l, tuple):
            l = list(l)
        elif not isinstance(l, list):
            l = [l, ]

        return l

    def gen_name():
        return 'pyodps_field_%s' % str(uuid.uuid4()).replace('-', '_')

    def _gen_actual_reducer(reducer, group):
        class ActualReducer(object):
            def __init__(self, resources=None):
                self._func = reducer
                self._curr = None
                self._prev_rows = None
                self._key_named_tuple = namedtuple('NamedKeys', group)

                self._resources = resources
                self._f = None

            def _is_generator_function(self, f):
                if inspect.isgeneratorfunction(f):
                    return True
                elif hasattr(f, '__call__') and inspect.isgeneratorfunction(f.__call__):
                    return True
                return False

            def __call__(self, row):
                key = tuple(getattr(row, n) for n in group)
                k = self._key_named_tuple(*key)

                if self._prev_rows is not None:
                    key_consumed = self._curr != key
                    if self._is_generator_function(self._f):
                        for it in self._f(self._prev_rows, key_consumed):
                            yield it
                    else:
                        res = self._f(self._prev_rows, key_consumed)
                        if res:
                            yield res

                self._prev_rows = row

                if self._curr is None or self._curr != key:
                    self._curr = key
                    if self._resources and self._f is None:
                        self._func = self._func(self._resources)
                    self._f = self._func(k)

            def close(self):
                if self._prev_rows and self._curr:
                    if self._is_generator_function(self._f):
                        for it in self._f(self._prev_rows, True):
                            yield it
                    else:
                        res = self._f(self._prev_rows, True)
                        if res:
                            yield res
                self._prev_rows = None

        return ActualReducer

    def _gen_combined_mapper(mapper, combiner, names, group, sort, ascending,
                             buffer_size, mapper_resources=None):
        mapper = mapper if not isinstance(mapper, FunctionWrapper) else mapper._func
        sort_indexes = [names.index(s) for s in sort]
        if isinstance(ascending, bool):
            ascending = [ascending] * len(sort)

        class CombinedMapper(object):
            def __init__(self, resources=None):
                if mapper_resources:
                    self.f = mapper(resources)
                elif inspect.isclass(mapper):
                    self.f = mapper()
                else:
                    self.f = mapper
                self.buffer = list()
                if inspect.isfunction(self.f):
                    self.is_generator = inspect.isgeneratorfunction(self.f)
                else:
                    self.is_generator = inspect.isgeneratorfunction(self.f.__call__)

            def _cmp_to_key(self, cmp):
                """Convert a cmp= function into a key= function"""

                class K(object):
                    def __init__(self, obj):
                        self.obj = obj

                    def __lt__(self, other):
                        return cmp(self.obj, other.obj) < 0

                    def __gt__(self, other):
                        return cmp(self.obj, other.obj) > 0

                    def __eq__(self, other):
                        return cmp(self.obj, other.obj) == 0

                    def __le__(self, other):
                        return cmp(self.obj, other.obj) <= 0

                    def __ge__(self, other):
                        return cmp(self.obj, other.obj) >= 0

                    def __ne__(self, other):
                        return cmp(self.obj, other.obj) != 0

                return K

            def _combine(self):
                def cmp(x, y):
                    for asc, sort_idx in zip(ascending, sort_indexes):
                        indict = 1 if asc else -1
                        if x[sort_idx] > y[sort_idx]:
                            return indict * 1
                        elif x[sort_idx] < y[sort_idx]:
                            return indict * -1
                        else:
                            continue

                    return 0

                self.buffer.sort(key=self._cmp_to_key(cmp))

                ActualCombiner = _gen_actual_reducer(combiner, group)
                ac = ActualCombiner()
                named_row = namedtuple('NamedRow', names)
                for r in self.buffer:
                    row = named_row(*r)
                    for l in ac(row):
                        yield l
                for l in ac.close():
                    yield l

                self.buffer = []

            def _handle_output_line(self, line):
                if len(self.buffer) >= buffer_size:
                    for l in self._combine():
                        yield l

                self.buffer.append(line)

            def __call__(self, row):
                if self.is_generator:
                    for it in self.f(row):
                        for l in self._handle_output_line(it):
                            yield l
                else:
                    for l in self._handle_output_line(self.f(row)):
                        yield l

            def close(self):
                if len(self.buffer) > 0:
                    for l in self._combine():
                        yield l

        return CombinedMapper

    mapper = _adjust_partial(mapper)
    reducer = _adjust_partial(reducer)
    combiner = _adjust_partial(combiner)

    if isinstance(mapper, FunctionWrapper):
        mapper_output_names = mapper_output_names or mapper.output_names
        mapper_output_types = mapper_output_types or mapper.output_types

    mapper_output_names = conv(mapper_output_names)
    mapper_output_types = conv(mapper_output_types)

    if mapper_output_types is not None and mapper_output_names is None:
        mapper_output_names = [gen_name() for _ in range(len(mapper_output_types))]

    if mapper is None and mapper_output_names is None:
        mapper_output_names = expr.schema.names

    group = conv(group) or mapper_output_names
    sort = sort or tuple()
    sort = list(OrderedDict.fromkeys(group + conv(sort)))
    if len(sort) > len(group):
        ascending = [ascending, ] * (len(sort) - len(group)) \
            if isinstance(ascending, bool) else list(ascending)
        if len(ascending) != len(sort):
            ascending = [True] * len(group) + ascending

    if not set(group + sort).issubset(mapper_output_names):
        raise ValueError('group and sort have to be the column names of mapper')

    if mapper is None:
        if mapper_output_names and mapper_output_names != expr.schema.names:
            raise ExpressionError(
                'Null mapper cannot have mapper output names: %s' % mapper_output_names)
        if mapper_output_types and mapper_output_types != expr.schema.types:
            raise ExpressionError(
                'Null mapper cannot have mapper output types: %s' % mapper_output_types)
        mapped = expr
        if combiner is not None:
            raise ValueError('Combiner is not null when mapper is null')
    else:
        if combiner is not None:
            if isinstance(combiner, FunctionWrapper):
                if combiner.output_names and \
                        combiner.output_names != mapper_output_names:
                    raise ExpressionError(
                        'Combiner must have the same output names with mapper')
                if combiner.output_types and \
                        combiner.output_types != mapper_output_types:
                    raise ExpressionError(
                        'Combiner must have the same output types with mapper')
                combiner = combiner._func
            mapper = _gen_combined_mapper(mapper, combiner, mapper_output_names,
                                          group, sort, ascending, combiner_buffer_size,
                                          mapper_resources=mapper_resources)
        mapped = expr.apply(mapper, axis=1, names=mapper_output_names,
                            types=mapper_output_types, resources=mapper_resources)

    clustered = mapped.groupby(group).sort(sort, ascending=ascending)

    if isinstance(reducer, FunctionWrapper):
        reducer_output_names = reducer_output_names or reducer.output_names
        reducer_output_types = reducer_output_types or reducer.output_types
        reducer = reducer._func

    if reducer is None:
        if reducer_output_names and reducer_output_names != mapped.schema.names:
            raise ExpressionError(
                'Null reducer cannot have reducer output names %s' % reducer_output_names)
        if reducer_output_types and reducer_output_types != mapped.schema.types:
            raise ExpressionError(
                'Null reducer cannot have reducer output types %s' % reducer_output_types)
        return mapped

    ActualReducer = _gen_actual_reducer(reducer, group)
    return clustered.apply(ActualReducer, resources=reducer_resources,
                           names=reducer_output_names, types=reducer_output_types)


class PivotCollectionExpr(DynamicCollectionExpr):
    _args = '_input', '_group', '_columns', '_values'
    node_name = 'Pivot'

    def _init(self, *args, **kwargs):
        self._init_attr('_group', None)
        self._init_attr('_columns', None)
        self._init_attr('_values', None)

        super(PivotCollectionExpr, self)._init(*args, **kwargs)

        if not hasattr(self, '_schema'):
            self._schema = DynamicSchema.from_lists(
                [f.name for f in self._group], [f.dtype for f in self._group]
            )

    def iter_args(self):
        for it in zip(['collection', 'group', 'columns', 'values'], self.args):
            yield it

    @property
    def input(self):
        return self._input

    def accept(self, visitor):
        return visitor.visit_pivot(self)


def pivot(expr, rows, columns, values=None):
    """
    Produce ‘pivot’ table based on 3 columns of this DataFrame.
    Uses unique values from rows / columns and fills with values.

    :param expr: collection
    :param rows: use to make new collection's grouped rows
    :param columns: use to make new collection's columns
    :param values: values to use for populating new collection's values
    :return: collection

    :Example:

    >>> df.pivot(rows='id', columns='category')
    >>> df.pivot(rows='id', columns='category', values='sale')
    >>> df.pivot(rows=['id', 'id2'], columns='category', values='sale')
    """

    rows = [expr._get_field(r) for r in utils.to_list(rows)]
    columns = [expr._get_field(c) for c in utils.to_list(columns)]
    if values:
        values = utils.to_list(values)
    else:
        names = set(c.name for c in rows + columns)
        values = [n for n in expr.schema.names if n not in names]
        if not values:
            raise ValueError('No values found for pivot')
    values = [expr._get_field(v) for v in values]

    if len(columns) > 1:
        raise ValueError('More than one `columns` are not supported yet')

    return PivotCollectionExpr(_input=expr, _group=rows,
                               _columns=columns, _values=values)


def melt(expr, id_vars=None, value_vars=None, var_name='variable', value_name='value', ignore_nan=False):
    """
    “Unpivots” a DataFrame from wide format to long format, optionally leaving identifier variables set.

    This function is useful to massage a DataFrame into a format where one or more columns are identifier
    variables (id_vars), while all other columns, considered measured variables (value_vars), are “unpivoted”
    to the row axis, leaving just two non-identifier columns, ‘variable’ and ‘value’.

    :param expr: collection
    :param id_vars: column(s) to use as identifier variables.
    :param value_vars: column(s) to unpivot. If not specified, uses all columns that are not set as id_vars.
    :param var_name: name to use for the ‘variable’ column. If None it uses frame.columns.name or ‘variable’.
    :param value_name: name to use for the ‘value’ column.
    :param ignore_nan: whether to ignore NaN values in data.
    :return: collection

    :Example:

    >>> df.melt(id_vars='id', value_vars=['col1', 'col2'])
    >>> df.melt(id_vars=['id', 'id2'], value_vars=['col1', 'col2'], var_name='variable')
    """
    id_vars = id_vars or []
    id_vars = [expr._get_field(r) for r in utils.to_list(id_vars)]
    if not value_vars:
        id_names = set([c.name for c in id_vars])
        value_vars = [expr._get_field(c) for c in expr.schema.names if c not in id_names]
    else:
        value_vars = [expr._get_field(c) for c in value_vars]

    col_type = utils.highest_precedence_data_type(*[c.dtype for c in value_vars])

    col_names = [c.name for c in value_vars]
    id_names = [r.name for r in id_vars]

    names = id_names + [var_name, value_name]
    dtypes = [r.dtype for r in id_vars] + [types.string, col_type]

    @output(names, dtypes)
    def mapper(row):
        for cn in col_names:
            col_value = getattr(row, cn)
            if ignore_nan and col_value is None:
                continue
            vals = [getattr(row, rn) for rn in id_names]
            yield tuple(vals + [cn, col_value])

    return expr.map_reduce(mapper)


class PivotTableCollectionExpr(CollectionExpr):
    __slots__ = '_agg_func', '_agg_func_names'
    _args = '_input', '_group', '_columns', '_values', '_fill_value'
    node_name = 'PivotTable'

    def _init(self, *args, **kwargs):
        for arg in self._args:
            self._init_attr(arg, None)

        super(PivotTableCollectionExpr, self)._init(*args, **kwargs)

        for attr in ('_fill_value', ):
            val = getattr(self, attr, None)
            if val is not None and not isinstance(val, Scalar):
                setattr(self, attr, Scalar(_value=val))

    @property
    def input(self):
        return self._input

    @property
    def fill_value(self):
        if self._fill_value:
            return self._fill_value.value

    @property
    def margins(self):
        return self._margins.value

    @property
    def margins_name(self):
        return self._margins_name.value

    def accept(self, visitor):
        return visitor.visit_pivot(self)


def pivot_table(expr, values=None, rows=None, columns=None, aggfunc='mean',
                fill_value=None):
    """
    Create a spreadsheet-style pivot table as a DataFrame.

    :param expr: collection
    :param values (optional): column to aggregate
    :param rows: rows to group
    :param columns: keys to group by on the pivot table column
    :param aggfunc: aggregate function or functions
    :param fill_value (optional): value to replace missing value with, default None
    :return: collection

    :Example:
    >>> df
        A    B      C   D
    0  foo  one  small  1
    1  foo  one  large  2
    2  foo  one  large  2
    3  foo  two  small  3
    4  foo  two  small  3
    5  bar  one  large  4
    6  bar  one  small  5
    7  bar  two  small  6
    8  bar  two  large  7
    >>> table = df.pivot_table(values='D', rows=['A', 'B'], columns='C', aggfunc='sum')
    >>> table
         A    B  large_D_sum   small_D_sum
    0  bar  one          4.0           5.0
    1  bar  two          7.0           6.0
    2  foo  one          4.0           1.0
    3  foo  two          NaN           6.0
    """

    def get_names(iters):
        return [r if isinstance(r, six.string_types) else r.name
                for r in iters]

    def get_aggfunc_name(f):
        if isinstance(f, six.string_types):
            if '(' in f:
                f = re.sub(r' *\( *', '_', f)
                f = re.sub(r' *[+\-\*/,] *', '_', f)
                f = re.sub(r' *\) *', '', f)
                f = f.replace('.', '_')
            return f
        if isinstance(f, FunctionWrapper):
            return f.output_names[0]
        return 'aggregation'

    if not rows:
        raise ValueError('No group keys passed')
    rows = utils.to_list(rows)
    rows_names = get_names(rows)
    rows = [expr._get_field(r) for r in rows]

    if isinstance(aggfunc, dict):
        agg_func_names = lkeys(aggfunc)
        aggfunc = lvalues(aggfunc)
    else:
        aggfunc = utils.to_list(aggfunc)
        agg_func_names = [get_aggfunc_name(af) for af in aggfunc]

    if not columns:
        if values is None:
            values = [n for n in expr.schema.names if n not in rows_names]
        else:
            values = utils.to_list(values)
        values = [expr._get_field(v) for v in values]

        names = rows_names
        types = [r.dtype for r in rows]
        for func, func_name in zip(aggfunc, agg_func_names):
            for value in values:
                if isinstance(func, six.string_types):
                    seq = value.eval(func, rewrite=False)
                    if isinstance(seq, ReprWrapper):
                        seq = seq()
                else:
                    seq = value.agg(func)
                seq = seq.rename('{0}_{1}'.format(value.name, func_name))
                names.append(seq.name)
                types.append(seq.dtype)
        schema = TableSchema.from_lists(names, types)

        return PivotTableCollectionExpr(_input=expr, _group=rows, _values=values,
                                        _fill_value=fill_value, _schema=schema,
                                        _agg_func=aggfunc, _agg_func_names=agg_func_names)
    else:
        columns = [expr._get_field(c) for c in utils.to_list(columns)]

        if values:
            values = utils.to_list(values)
        else:
            names = set(c.name for c in rows + columns)
            values = [n for n in expr.schema.names if n not in names]
            if not values:
                raise ValueError('No values found for pivot_table')
        values = [expr._get_field(v) for v in values]

        if len(columns) > 1:
            raise ValueError('More than one `columns` are not supported yet')

        schema = DynamicSchema.from_lists(rows_names, [r.dtype for r in rows])
        base_tp = PivotTableCollectionExpr
        tp = type(base_tp.__name__, (DynamicCollectionExpr, base_tp), dict())
        return tp(_input=expr, _group=rows, _values=values,
                  _columns=columns, _agg_func=aggfunc,
                  _fill_value=fill_value, _schema=schema,
                  _agg_func_names=agg_func_names)


def _scale_values(expr, columns, agg_fun, scale_fun, preserve=False, suffix='_scaled', group=None):
    from ..types import Float, Integer
    time_suffix = str(int(time.time()))

    if group is not None:
        group = utils.to_list(group)
        group = [expr._get_field(c).name if isinstance(c, Column) else c for c in group]

    if columns is None:
        if group is None:
            columns = expr.schema.names
        else:
            columns = [n for n in expr.schema.names if n not in group]
    else:
        columns = utils.to_list(columns)
    columns = [expr._get_field(v) for v in columns]

    numerical_cols = [col.name for col in columns if isinstance(col.data_type, (Float, Integer))]

    agg_cols = []
    for col_name in numerical_cols:
        agg_cols.extend(agg_fun(expr, col_name))

    if group is None:
        # make a fake constant column to join
        extra_col = 'idx_col_' + time_suffix
        join_cols = [extra_col]
        stats_df = expr.__getitem__([Scalar(1).rename(extra_col)] + agg_cols)
        mapped = expr[expr, Scalar(1).rename(extra_col)]
    else:
        extra_col = None
        join_cols = group
        stats_df = expr.groupby(join_cols).agg(*agg_cols)
        mapped = expr

    joined = mapped.join(stats_df, on=join_cols, mapjoin=True)
    if extra_col is not None:
        joined = joined.exclude(extra_col)

    if preserve:
        norm_cols = [dt.name for dt in expr.dtypes]
        norm_cols.extend([scale_fun(joined, dt.name).rename(dt.name + suffix)
                          for dt in expr.dtypes if dt.name in numerical_cols])
    else:
        norm_cols = [scale_fun(joined, dt.name).rename(dt.name)
                     if dt.name in numerical_cols else getattr(joined, dt.name)
                     for dt in expr.dtypes]
    return joined.__getitem__(norm_cols)


def min_max_scale(expr, columns=None, feature_range=(0, 1), preserve=False, suffix='_scaled', group=None):
    """
    Resize a data frame by max / min values, i.e., (X - min(X)) / (max(X) - min(X))

    :param DataFrame expr: input DataFrame
    :param feature_range: the target range to resize the value into, i.e., v * (b - a) + a
    :param bool preserve: determine whether input data should be kept. If True, scaled input data will be appended to the data frame with `suffix`
    :param columns: columns names to resize. If set to None, float or int-typed columns will be normalized if the column is not specified as a group column.
    :param group: determine scale groups. Scaling will be done in each group separately.
    :param str suffix: column suffix to be appended to the scaled columns.

    :return: resized data frame
    :rtype: DataFrame
    """
    time_suffix = str(int(time.time()))

    def calc_agg(expr, col):
        return [
            getattr(expr, col).min().rename(col + '_min_' + time_suffix),
            getattr(expr, col).max().rename(col + '_max_' + time_suffix),
        ]

    def do_scale(expr, col):
        f_min, f_max = feature_range
        r = getattr(expr, col + '_max_' + time_suffix) - getattr(expr, col + '_min_' + time_suffix)
        scaled = (r == 0).ifelse(Scalar(0), (getattr(expr, col) - getattr(expr, col + '_min_' + time_suffix)) / r)
        return scaled * (f_max - f_min) + f_min

    return _scale_values(expr, columns, calc_agg, do_scale, preserve=preserve, suffix=suffix, group=group)


def std_scale(expr, columns=None, with_means=True, with_std=True, preserve=False, suffix='_scaled', group=None):
    """
    Resize a data frame by mean and standard error.

    :param DataFrame expr: Input DataFrame
    :param bool with_means: Determine whether the output will be subtracted by means
    :param bool with_std: Determine whether the output will be divided by standard deviations
    :param bool preserve: Determine whether input data should be kept. If True, scaled input data will be appended to the data frame with `suffix`
    :param columns: Columns names to resize. If set to None, float or int-typed columns will be normalized if the column is not specified as a group column.
    :param group: determine scale groups. Scaling will be done in each group separately.
    :param str suffix: column suffix to be appended to the scaled columns.

    :return: resized data frame
    :rtype: DataFrame
    """
    time_suffix = str(int(time.time()))

    def calc_agg(expr, col):
        return [
            getattr(expr, col).mean().rename(col + '_mean_' + time_suffix),
            getattr(expr, col).std(ddof=0).rename(col + '_std_' + time_suffix),
        ]

    def do_scale(expr, col):
        c = getattr(expr, col)
        mean_expr = getattr(expr, col + '_mean_' + time_suffix)
        if with_means:
            c = c - mean_expr
            mean_expr = Scalar(0)
        if with_std:
            std_expr = getattr(expr, col + '_std_' + time_suffix)
            c = (std_expr == 0).ifelse(mean_expr, c / getattr(expr, col + '_std_' + time_suffix))
        return c

    return _scale_values(expr, columns, calc_agg, do_scale, preserve=preserve, suffix=suffix, group=group)


class ExtractKVCollectionExpr(DynamicCollectionExpr):
    __slots__ = '_column_type',
    _args = '_input', '_columns', '_intact', '_kv_delimiter', '_item_delimiter', '_default'
    node_name = 'ExtractKV'

    def _init(self, *args, **kwargs):
        from .element import _scalar
        for attr in self._args[1:]:
            self._init_attr(attr, None)
        super(ExtractKVCollectionExpr, self)._init(*args, **kwargs)
        self._kv_delimiter = _scalar(self._kv_delimiter)
        self._item_delimiter = _scalar(self._item_delimiter)
        self._default = _scalar(self._default)

    @property
    def input(self):
        return self._input

    def accept(self, visitor):
        visitor.visit_extract_kv(self)


def extract_kv(expr, columns=None, kv_delim=':', item_delim=',', dtype='float', fill_value=None):
    """
    Extract values in key-value represented columns into standalone columns. New column names will
    be the name of the key-value column followed by an underscore and the key.

    :param DataFrame expr: input DataFrame
    :param columns: the key-value columns to be extracted.
    :param str kv_delim: delimiter between key and value.
    :param str item_delim: delimiter between key-value pairs.
    :param str dtype: type of value columns to generate.
    :param fill_value: default value for missing key-value pairs.

    :return: extracted data frame
    :rtype: DataFrame

    :Example:
    >>> df
        name   kv
    0  name1  k1=1.0,k2=3.0,k5=10.0
    1  name2  k2=3.0,k3=5.1
    2  name3  k1=7.1,k7=8.2
    3  name4  k2=1.2,k3=1.5
    4  name5  k2=1.0,k9=1.1
    >>> table = df.extract_kv(columns=['A', 'B'], kv_delim='=')
    >>> table
        name   kv_k1   kv_k2   kv_k3   kv_k5   kv_k7   kv_k9
    0  name1  1.0     3.0     Nan     10.0    Nan     Nan
    1  name2  Nan     3.0     5.1     Nan     Nan     Nan
    2  name3  7.1     Nan     Nan     Nan     8.2     Nan
    3  name4  Nan     1.2     1.5     Nan     Nan     Nan
    4  name5  Nan     1.0     Nan     Nan     Nan     1.1
    """
    if columns is None:
        columns = [expr._get_field(c) for c in expr.schema.names]
        intact_cols = []
    else:
        columns = [expr._get_field(c) for c in utils.to_list(columns)]
        name_set = set([c.name for c in columns])
        intact_cols = [expr._get_field(c) for c in expr.schema.names if c not in name_set]

    column_type = types.validate_data_type(dtype)
    if any(not isinstance(c.dtype, types.String) for c in columns):
        raise ExpressionError('Key-value columns must be strings.')

    schema = DynamicSchema.from_lists([c.name for c in intact_cols], [c.dtype for c in intact_cols])
    return ExtractKVCollectionExpr(_input=expr, _columns=columns, _intact=intact_cols, _schema=schema,
                                   _column_type=column_type, _default=fill_value,
                                   _kv_delimiter=kv_delim, _item_delimiter=item_delim)


def to_kv(expr, columns=None, kv_delim=':', item_delim=',', kv_name='kv_col'):
    """
    Merge values in specified columns into a key-value represented column.

    :param DataFrame expr: input DataFrame
    :param columns: the columns to be merged.
    :param str kv_delim: delimiter between key and value.
    :param str item_delim: delimiter between key-value pairs.
    :param str kv_col: name of the new key-value column

    :return: converted data frame
    :rtype: DataFrame

    :Example:
    >>> df
        name   k1   k2   k3   k5    k7   k9
    0  name1  1.0  3.0  Nan  10.0  Nan  Nan
    1  name2  Nan  3.0  5.1  Nan   Nan  Nan
    2  name3  7.1  Nan  Nan  Nan   8.2  Nan
    3  name4  Nan  1.2  1.5  Nan   Nan  Nan
    4  name5  Nan  1.0  Nan  Nan   Nan  1.1
    >>> table = df.to_kv(columns=['A', 'B'], kv_delim='=')
    >>> table
        name   kv_col
    0  name1  k1=1.0,k2=3.0,k5=10.0
    1  name2  k2=3.0,k3=5.1
    2  name3  k1=7.1,k7=8.2
    3  name4  k2=1.2,k3=1.5
    4  name5  k2=1.0,k9=1.1
    """
    if columns is None:
        columns = [expr._get_field(c) for c in expr.schema.names]
        intact_cols = []
    else:
        columns = [expr._get_field(c) for c in utils.to_list(columns)]
        name_set = set([c.name for c in columns])
        intact_cols = [expr._get_field(c) for c in expr.schema.names if c not in name_set]

    mapped_cols = [c.isnull().ifelse(Scalar(''), c.name + kv_delim + c.astype('string')) for c in columns]
    reduced_col = reduce(lambda a, b: (b == '').ifelse(a, (a == '').ifelse(b, a + item_delim + b)), mapped_cols)
    return expr.__getitem__(intact_cols + [reduced_col.rename(kv_name)])


def dropna(expr, how='any', thresh=None, subset=None):
    """
    Return object with labels on given axis omitted where alternately any or all of the data are missing

    :param DataFrame expr: input DataFrame
    :param how: can be ‘any’ or ‘all’. If 'any' is specified any NA values are present, drop that label. If 'all' is specified and all values are NA, drop that label.
    :param thresh: require that many non-NA values
    :param subset: Labels along other axis to consider, e.g. if you are dropping rows these would be a list of columns to include
    :return: DataFrame
    """
    if subset is None:
        subset = [expr._get_field(c) for c in expr.schema.names]
    else:
        subset = [expr._get_field(c) for c in utils.to_list(subset)]

    if not subset:
        raise ValueError('Illegal subset is provided.')

    if thresh is None:
        thresh = len(subset) if how == 'any' else 1

    sum_exprs = reduce(operator.add, (s.notna().ifelse(1, 0) for s in subset))
    return expr.filter(sum_exprs >= thresh)


def fillna(expr, value=None, method=None, subset=None):
    """
    Fill NA/NaN values using the specified method

    :param DataFrame expr: input DataFrame
    :param method: can be ‘backfill’, ‘bfill’, ‘pad’, ‘ffill’ or None
    :param value: value to fill into
    :param subset: Labels along other axis to consider.
    :return: DataFrame
    """
    col_dict = OrderedDict([(c, expr._get_field(c)) for c in expr.schema.names])
    if subset is None:
        sel_col_names = expr.schema.names
    else:
        # when c is in expr._fields, _get_field may do substitution which will cause error
        subset = (c.copy() if isinstance(c, Expr) else c for c in utils.to_list(subset))
        sel_col_names = [expr._get_field(c).name for c in subset]

    if method is not None and value is not None:
        raise ValueError('The argument `method` is not compatible with `value`.')
    if method is None and value is None:
        raise ValueError('You should supply at least one argument in `method` and `value`.')
    if method is not None and method not in ('backfill', 'bfill', 'pad', 'ffill'):
        raise ValueError('Method value %s is illegal.' % str(method))

    if method in ('backfill', 'bfill'):
        sel_cols = list(reversed(sel_col_names))
    else:
        sel_cols = sel_col_names

    if method is None:
        for n in sel_col_names:
            e = col_dict[n]
            col_dict[n] = e.isna().ifelse(value, e).rename(n)
        return expr.select(list(col_dict.values()))

    else:
        names = list(col_dict.keys())
        typs = list(c.dtype.name for c in col_dict.values())

        @output(names, typs)
        def mapper(row):
            last_valid = None
            update_dict = dict()

            import math
            try:
                import numpy as np
            except ImportError:
                np = None

            def isnan(v):
                if v is None:
                    return True
                if np is not None:
                    try:
                        return np.isnan(v)
                    except TypeError:
                        pass
                try:
                    return math.isnan(v)
                except TypeError:
                    return False

            for n in sel_cols:
                old_val = getattr(row, n)
                if old_val is None or isnan(old_val):
                    if last_valid is not None:
                        update_dict[n] = last_valid
                else:
                    last_valid = old_val

            yield row.replace(**update_dict)

        return expr.map_reduce(mapper)


def ffill(expr, subset=None):
    """
    Fill NA/NaN values with the forward method. Equivalent to fillna(method='ffill').

    :param DataFrame expr: input DataFrame.
    :param subset: Labels along other axis to consider.
    :return: DataFrame
    """
    return expr.fillna(method='ffill', subset=subset)


def bfill(expr, subset=None):
    """
    Fill NA/NaN values with the backward method. Equivalent to fillna(method='bfill').

    :param DataFrame expr: input DataFrame.
    :param subset: Labels along other axis to consider.
    :return: DataFrame
    """
    return expr.fillna(method='bfill', subset=subset)


class AppendIDCollectionExpr(CollectionExpr):
    _args = '_input', '_id_col'
    node_name = 'AppendID'

    def _init(self, *args, **kwargs):
        from .element import _scalar
        for attr in self._args[1:]:
            self._init_attr(attr, None)
        super(AppendIDCollectionExpr, self)._init(*args, **kwargs)
        self._validate()
        self._id_col = _scalar(self._id_col)
        self._schema = TableSchema.from_lists(
            self._input.schema.names + [self._id_col.value],
            self._input.schema.types + [types.int64],
        )

    def _validate(self):
        if self._id_col in self._input.schema:
            raise ExpressionError('ID column already exists in current data frame.')

    @property
    def input(self):
        return self._input

    def accept(self, visitor):
        return visitor.visit_append_id(self)


def _append_id(expr, id_col='append_id'):
    return AppendIDCollectionExpr(_input=expr, _id_col=id_col)


def append_id(expr, id_col='append_id'):
    """
    Append an ID column to current column to form a new DataFrame.

    :param str id_col: name of appended ID field.

    :return: DataFrame with ID field
    :rtype: DataFrame
    """
    if hasattr(expr, '_xflow_append_id'):
        return expr._xflow_append_id(id_col)
    else:
        return _append_id(expr, id_col)


class SplitCollectionExpr(CollectionExpr):
    _args = '_input', '_frac', '_seed', '_split_id'
    node_name = 'Split'

    def _init(self, *args, **kwargs):
        from .element import _scalar
        for attr in self._args[1:]:
            self._init_attr(attr, None)
        super(SplitCollectionExpr, self)._init(*args, **kwargs)
        self._frac = _scalar(self._frac)
        self._seed = _scalar(self._seed, types.int32)
        self._split_id = _scalar(self._split_id, types.int32)
        self._schema = self._input.schema

    @property
    def input(self):
        return self._input

    def accept(self, visitor):
        return visitor.visit_split(self)


def _split(expr, frac, seed=None):
    seed = seed or int(time.time())
    return (
        SplitCollectionExpr(_input=expr, _frac=frac, _seed=seed, _split_id=0),
        SplitCollectionExpr(_input=expr, _frac=frac, _seed=seed, _split_id=1),
    )


def split(expr, frac, seed=None):
    """
    Split the current column into two column objects with certain ratio.

    :param float frac: Split ratio

    :return: two split DataFrame objects
    """
    if hasattr(expr, '_xflow_split'):
        return expr._xflow_split(frac, seed=seed)
    else:
        return _split(expr, frac, seed=seed)


def applymap(expr, func, rtype=None, resources=None, columns=None, excludes=None, args=(), **kwargs):
    """
    Call func on each element of this collection.

    :param func: lambda, function, :class:`odps.models.Function`,
                 or str which is the name of :class:`odps.models.Funtion`
    :param rtype: if not provided, will be the dtype of this sequence
    :param columns: columns to apply this function on
    :param excludes: columns to skip when applying the function
    :return: a new collection

    :Example:

    >>> df.applymap(lambda x: x + 1)
    """
    if columns is not None and excludes is not None:
        raise ValueError('`columns` and `excludes` cannot be provided at the same time.')
    if not columns:
        excludes = excludes or []
        if isinstance(excludes, six.string_types):
            excludes = [excludes]
        excludes = set([c if isinstance(c, six.string_types) else c.name for c in excludes])
        columns = set([c for c in expr.schema.names if c not in excludes])
    else:
        if isinstance(columns, six.string_types):
            columns = [columns]
        columns = set([c if isinstance(c, six.string_types) else c.name for c in columns])
    mapping = [expr[c] if c not in columns
               else expr[c].map(func, rtype=rtype, resources=resources, args=args, **kwargs)
               for c in expr.schema.names]
    return expr.select(*mapping)


_collection_methods = dict(
    sort_values=sort_values,
    sort=sort_values,
    distinct=distinct,
    apply=apply,
    reshuffle=reshuffle,
    map_reduce=map_reduce,
    sample=sample,
    __sample=__df_sample,
    pivot=pivot,
    melt=melt,
    pivot_table=pivot_table,
    extract_kv=extract_kv,
    to_kv=to_kv,
    dropna=dropna,
    fillna=fillna,
    ffill=ffill,
    bfill=bfill,
    min_max_scale=min_max_scale,
    std_scale=std_scale,
    _append_id=_append_id,
    append_id=append_id,
    _split=_split,
    split=split,
    applymap=applymap,
)

_sequence_methods = dict(
    unique=unique
)

utils.add_method(CollectionExpr, _collection_methods)
utils.add_method(SequenceExpr, _sequence_methods)



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Client for interacting with the MaxCompute Storage API."""

import collections
import json
import logging
from io import IOBase, BytesIO
from enum import Enum
from hashlib import md5

try:
    import pyarrow as pa
except ImportError:
    pa = None
from requests import codes

from ... import ODPS, options, serializers
from ...models import Table
from ...models.core import JSONRemoteModel
from ...tunnel.io import RequestsIO
from ...utils import to_binary

STORAGE_VERSION = "1"
URL_PREFIX = "/api/storage/v" + STORAGE_VERSION

logger = logging.getLogger(__name__)


class Status(Enum):
    INIT = "INIT"
    OK = "OK"
    WAIT = "WAIT"
    RUNNING = "RUNNING"


class SessionStatus(Enum):
    INIT = "INIT"
    NORMAL = "NORMAL"
    CRITICAL = "CRITICAL"
    EXPIRED = "EXPIRED"
    COMMITTING = "COMMITTING"
    COMMITTED = "COMMITTED"


class SplitOptions(JSONRemoteModel):
    class SplitMode(str, Enum):
        SIZE = "Size"
        PARALLELISM = "Parallelism"
        ROW_OFFSET = "RowOffset"
        BUCKET = "Bucket"

    split_mode = serializers.JSONNodeField('SplitMode', parse_callback=lambda s: SplitOptions.SplitMode(s))
    split_number = serializers.JSONNodeField("SplitNumber")
    cross_partition = serializers.JSONNodeField("CrossPartition")

    def __init__(self, **kwargs):
        super(SplitOptions, self).__init__(**kwargs)

        self.split_mode = self.split_mode or SplitOptions.SplitMode.SIZE
        self.split_number = self.split_number or 256*1024*1024
        self.cross_partition = self.cross_partition or True

    @classmethod
    def get_default_options(self, mode):
        options = SplitOptions()
        options.cross_partition = True
        if mode == SplitOptions.SplitMode.SIZE:
            options.split_mode = SplitOptions.SplitMode.SIZE
            options.split_number = 256*1024*1024
        elif mode == SplitOptions.SplitMode.PARALLELISM:
            options.split_mode = SplitOptions.SplitMode.PARALLELISM
            options.split_number = 32
        elif mode == SplitOptions.SplitMode.ROW_OFFSET:
            options.split_mode = SplitOptions.SplitMode.ROW_OFFSET
            options.split_number = 0
        elif mode == SplitOptions.SplitMode.BUCKET:
            options.split_mode = SplitOptions.SplitMode.BUCKET

        return options


class ArrowOptions(JSONRemoteModel):
    class TimestampUnit(str, Enum):
        SECOND = "second"
        MILLI = "milli"
        MICRO = "micro"
        NANO = "nano"

    timestamp_unit = serializers.JSONNodeField('TimestampUnit', parse_callback=lambda s: ArrowOptions.TimestampUnit(s))
    date_time_unit = serializers.JSONNodeField('DatetimeUnit', parse_callback=lambda s: ArrowOptions.TimestampUnit(s))

    def __init__(self, **kwargs):
        super(ArrowOptions, self).__init__(**kwargs)

        self.timestamp_unit = self.timestamp_unit or ArrowOptions.TimestampUnit.NANO
        self.date_time_unit = self.date_time_unit or ArrowOptions.TimestampUnit.MILLI


class Column(JSONRemoteModel):
    name = serializers.JSONNodeField("Name")
    type = serializers.JSONNodeField("Type")
    comment = serializers.JSONNodeField("Comment")
    nullable = serializers.JSONNodeField("Nullable")


class DataSchema(JSONRemoteModel):
    data_columns = serializers.JSONNodesReferencesField(Column, 'DataColumns')
    partition_columns = serializers.JSONNodesReferencesField(Column, 'PartitionColumns')


class DataFormat(JSONRemoteModel):
    type = serializers.JSONNodeField("Type")
    version = serializers.JSONNodeField("Version")


class DynamicPartitionOptions(JSONRemoteModel):
    invalid_strategy = serializers.JSONNodeField("InvalidStrategy")
    invalid_limit = serializers.JSONNodeField("InvalidLimit")
    dynamic_partition_limit = serializers.JSONNodeField("DynamicPartitionLimit")

    def __init__(self, **kwargs):
        super(DynamicPartitionOptions, self).__init__(**kwargs)

        self.invalid_strategy = self.invalid_strategy or "Exception"
        self.invalid_limit = self.invalid_limit or 1
        self.dynamic_partition_limit = self.dynamic_partition_limit or 512


class Order(JSONRemoteModel):
    name = serializers.JSONNodeField("Name")
    sort_direction = serializers.JSONNodeField("SortDirection")


class RequiredDistribution(JSONRemoteModel):
    type = serializers.JSONNodeField("Type")
    cluster_keys = serializers.JSONNodeField("ClusterKeys")
    buckets_number = serializers.JSONNodeField("BucketsNumber")


class Compression(Enum):
    UNCOMPRESSED = 0
    ZSTD = 1
    LZ4_FRAME = 2

    def to_string(self):
        if self.value == 0:
            return None
        elif self.value == 1:
            return "zstd"
        elif self.value == 2:
            return "lz4"
        else:
            return "unknown"


class TableBatchScanRequest(serializers.JSONSerializableModel):
    required_data_columns = serializers.JSONNodeField("RequiredDataColumns")
    required_partition_columns = serializers.JSONNodeField("RequiredPartitionColumns")
    required_partitions = serializers.JSONNodeField("RequiredPartitions")
    required_bucket_ids = serializers.JSONNodeField("RequiredBucketIds")
    split_options = serializers.JSONNodeReferenceField(SplitOptions, "SplitOptions")
    arrow_options = serializers.JSONNodeReferenceField(ArrowOptions, "ArrowOptions")
    filter_predicate = serializers.JSONNodeField("FilterPredicate")

    def __init__(self, **kwargs):
        super(TableBatchScanRequest, self).__init__(**kwargs)

        self.required_data_columns = self.required_data_columns or []
        self.required_partition_columns = self.required_partition_columns or []
        self.required_partitions = self.required_partitions or []
        self.required_bucket_ids = self.required_bucket_ids or []
        self.split_options = self.split_options or SplitOptions()
        self.arrow_options = self.arrow_options or ArrowOptions()
        self.filter_predicate = self.filter_predicate or ""


class TableBatchScanResponse(serializers.JSONSerializableModel):
    __slots__ = ['status', 'request_id']

    session_id = serializers.JSONNodeField("SessionId")
    session_type = serializers.JSONNodeField("SessionType")
    session_status = serializers.JSONNodeField('SessionStatus', parse_callback=lambda s: SessionStatus(s.upper()))
    expiration_time = serializers.JSONNodeField("ExpirationTime")
    split_count = serializers.JSONNodeField("SplitsCount")
    record_count = serializers.JSONNodeField("RecordCount")
    data_schema = serializers.JSONNodeReferenceField(DataSchema, 'DataSchema')
    supported_data_format = serializers.JSONNodesReferencesField(DataFormat, "SupportedDataFormat")

    def __init__(self):
        super(TableBatchScanResponse, self).__init__()

        self.status = Status.INIT
        self.request_id = ""


class SessionRequest(object):
    def __init__(self, session_id):
        self.session_id = session_id


class TableBatchWriteRequest(serializers.JSONSerializableModel):
    dynamic_partition_options = serializers.JSONNodeReferenceField(DynamicPartitionOptions, "DynamicPartitionOptions")
    arrow_options = serializers.JSONNodeReferenceField(ArrowOptions, "ArrowOptions")
    overwrite = serializers.JSONNodeField("Overwrite")
    partition_spec = serializers.JSONNodeField("PartitionSpec")
    support_write_cluster = serializers.JSONNodeField("SupportWriteCluster")

    def __init__(self, **kwargs):
        super(TableBatchWriteRequest, self).__init__(**kwargs)

        self.partition_spec = self.partition_spec or ""
        self.arrow_options = self.arrow_options or ArrowOptions()
        self.dynamic_partition_options = self.dynamic_partition_options or DynamicPartitionOptions()
        self.overwrite = self.overwrite or True
        self.support_write_cluster = self.support_write_cluster or False


class TableBatchWriteResponse(serializers.JSONSerializableModel):
    __slots__ = ['status', "request_id"]

    session_status = serializers.JSONNodeField('SessionStatus', parse_callback=lambda s: SessionStatus(s.upper()))
    expiration_time = serializers.JSONNodeField("ExpirationTime")
    session_id = serializers.JSONNodeField("SessionId")
    data_schema = serializers.JSONNodeReferenceField(DataSchema, "DataSchema")
    supported_data_format = serializers.JSONNodesReferencesField(DataFormat, "SupportedDataFormat")
    max_block_num = serializers.JSONNodeField("MaxBlockNumber")
    required_ordering = serializers.JSONNodesReferencesField(Order, "RequiredOrdering")
    required_distribution = serializers.JSONNodeReferenceField(RequiredDistribution, "RequiredDistribution")

    def __init__(self):
        super(TableBatchWriteResponse, self).__init__()

        self.status = Status.INIT
        self.request_id = ""


class ReadRowsRequest(object):
    def __init__(self, session_id, split_index=0,
                 row_index=0, row_count=0, max_batch_rows=4096,
                 compression=Compression.LZ4_FRAME,
                 data_format=DataFormat()):
        self.session_id = session_id
        self.split_index = split_index
        self.row_index = row_index
        self.row_count = row_count
        self.max_batch_rows = max_batch_rows
        self.compression = compression
        self.data_format = data_format


class ReadRowsResponse(object):
    def __init__(self):
        self.status = Status.INIT
        self.request_id = ""


class WriteRowsRequest(object):
    def __init__(self, session_id,
                 block_number=0, attempt_number=0, bucket_id=0,
                 compression=Compression.LZ4_FRAME,
                 data_format=DataFormat()):
        self.session_id = session_id
        self.block_number = block_number
        self.attempt_number = attempt_number
        self.bucket_id = bucket_id
        self.compression = compression
        self.data_format = data_format


class WriteRowsResponse(object):
    def __init__(self):
        self.status = Status.INIT
        self.request_id = ""
        self.commit_message = ""


def update_request_id(response, resp):
    if "x-odps-request-id" in resp.headers:
        response.request_id = resp.headers["x-odps-request-id"]


class StreamReader(IOBase):
    """Stream reader."""

    def __init__(self, download):
        self._stopped = False
        raw_reader = download()

        self._raw_reader = raw_reader
        # need to confirm read size
        self._chunk_size = 65536
        self._buffers = collections.deque()

    def readable(self):
        """Check whether this stream reader has been closed or not.

        Returns:
            Readable or not.
        """
        return not self._stopped

    def _read_chunk(self):
        buf = self._raw_reader.raw.read(self._chunk_size)
        return buf

    def _fill_next_buffer(self):
        data = self._read_chunk()
        if len(data) == 0:
            return

        self._buffers.append(BytesIO(data))

    def read(self, nbytes=None):
        """Read stream data from the server.

        Args:
            nbytes: The number of bytes to be read. All data will be read at once if set to None.

        Returns:
            Stream data. None means all the data has been read or there is error occurred.
        """
        if self._stopped:
            return b''

        total_size = 0
        bufs = []
        while nbytes is None or total_size < nbytes:
            if not self._buffers:
                self._fill_next_buffer()
                if not self._buffers:
                    break

            to_read = nbytes - total_size if nbytes is not None else None
            buf = self._buffers[0].read(to_read)

            if not buf:
                self._buffers.popleft()
            else:
                bufs.append(buf)
                total_size += len(buf)

        return b''.join(bufs)

    def get_status(self):
        """Get the status of the stream reader.

        Returns:
            Status.OK or Status.RUNNING.
        """
        if not self._stopped:
            return Status.RUNNING
        else:
            return Status.OK

    def get_request_id(self):
        """Get the request id.

        Returns:
            Request id.
        """
        if not self._stopped:
            logger.error("The reader is not closed yet, please wait")
            return None

        if self._raw_reader is not None and "x-odps-request-id" in self._raw_reader.headers:
            return self._raw_reader.headers["x-odps-request-id"]
        else:
            return None

    def close(self):
        """If there is no data can be read from server, it will be called to close the stream reader."""
        self._stopped = True


class StreamWriter(IOBase):
    """Stream writer."""

    def __init__(self, upload):
        self._req_io = RequestsIO(upload, chunk_size=options.chunk_size)
        self._req_io.start()
        self._res = None
        self._stopped = False

    def writable(self):
        """Check whether this stream writer has been closed or not.

        Returns:
            Writable or not.
        """
        return not self._stopped

    def write(self, data):
        """Write data to the server.

        Returns:
            Success or not.
        """
        if self._stopped:
            return False

        self._req_io.write(data)
        return True

    def finish(self):
        """The stream writer is not expected to write data if finish has been called.

        Returns:
            Commit message returned from the server. User should bring this message to do the write session commit.
            Success or not.
        """
        self._stopped = True
        self._res = self._req_io.finish()

        if self._res is not None and self._res.status_code == codes['ok']:
            resp_json = self._res.json()
            return resp_json["CommitMessage"], True
        else:
           return None, False

    def get_status(self):
        """Get the status of this stream writer.

        Returns:
            Status.OK or Status.RUNNING.
        """
        if not self._stopped:
            return Status.RUNNING
        else:
            return Status.OK

    def get_request_id(self):
        """Get the request id.

        Returns:
            Request id.
        """
        if not self._stopped:
            logger.error("The writer is not closed yet, please close first")
            return None

        if self._res is not None and "x-odps-request-id" in self._res.headers:
            return self._res.headers["x-odps-request-id"]
        else:
            return None


class ArrowReader(object):
    """Arrow batch reader."""

    def __init__(self, stream_reader):
        if pa is None:
            raise ValueError("To use arrow reader you need to install pyarrow")

        self._reader = stream_reader
        self._arrow_stream = None

    def _read_next_batch(self):
        if self._arrow_stream is None:
            self._arrow_stream = pa.ipc.open_stream(self._reader)

        try:
            batch = self._arrow_stream.read_next_batch()
            return batch
        except StopIteration:
            return None

    def read(self):
        """Read arrow batch from the server.

        Returns:
            Arrow record batch. None means all the data has been read or there is error occurred.
        """
        if not self._reader.readable():
            logger.error("Reader has been closed")
            return None

        batch = self._read_next_batch()
        if batch is None:
            self._reader.close()

        return batch

    def get_status(self):
        """Get the status of the arrow batch reader.

        Returns:
            Status.OK or Status.RUNNING.
        """
        return self._reader.get_status()

    def get_request_id(self):
        """Get the request id.

        Returns:
            Request id.
        """
        return self._reader.get_request_id()


class ArrowWriter(object):
    """Arrow batch writer."""

    def __init__(self, stream_writer, compression):
        self._arrow_writer = None
        self._compression = compression
        self._sink = stream_writer

    def write(self, record_batch):
        """Write one arrow batch to the server.

        Args:
            record_batch: The arrow batch to be written.
        Returns:
            Success or not.
        """
        if not self._sink.writable():
            logger.error("Writer has been closed")
            return False

        if self._arrow_writer is None:
            self._arrow_writer = pa.ipc.new_stream(
                self._sink,
                record_batch.schema,
                options=pa.ipc.IpcWriteOptions(compression=self._compression.to_string()),
            )

        self._arrow_writer.write_batch(record_batch)

        if not self._sink.writable():
            logger.error("Writer has been closed as exception occurred")
            return False

        return True

    def finish(self):
        """The arrow writer is not expected to write data if finish has been called.

        Returns:
            Commit message returned from the server. User should bring this message
            to do the write session commit.
            Success ot not.
        """
        if self._arrow_writer:
            self._arrow_writer.close()
        return self._sink.finish()

    def get_status(self):
        """Get the status of the arrow batch writer.

        Returns:
            Status.OK or Status.RUNNING.
        """
        return self._sink.get_status()

    def get_request_id(self):
        """Get the request id.

        Returns:
            Request id.
        """
        return self._sink.get_request_id()


class StorageApiClient(object):
    """Client to bundle configuration needed for API requests."""

    def __init__(self, odps: ODPS, table: Table, rest_endpoint: str = None, quota_name: str = None):
        if isinstance(odps, ODPS) and isinstance(table, Table):
            self._odps = odps
            self._table = table
            self._quota_name = quota_name
            self._rest_endpoint = rest_endpoint
            self._tunnel_rest = None
        else:
            raise ValueError("Please input odps configuration")

    @property
    def table(self):
        return self._table

    @property
    def tunnel_rest(self):
        if self._tunnel_rest is not None:
            return self._tunnel_rest

        from ...tunnel.tabletunnel import TableTunnel

        tunnel = TableTunnel(self._odps, endpoint=self._rest_endpoint, quota_name=self._quota_name)
        self._tunnel_rest = tunnel.tunnel_rest
        return self._tunnel_rest

    def _get_resource(self, *args) -> str:
        endpoint = self.tunnel_rest.endpoint + URL_PREFIX
        url = self._table.table_resource(endpoint=endpoint, force_schema=True)
        return "/".join([url] + list(args))

    def create_read_session(self, request: TableBatchScanRequest) -> TableBatchScanResponse:
        """Create a read session.

        Args:
            request: Table split parameters sent to the server.

        Returns:
            Read session response returned from the server.
        """
        if not isinstance(request, TableBatchScanRequest):
            raise ValueError("Use TableBatchScanRequest class to build request for create read session interface")

        json_str = request.serialize()

        url = self._get_resource("sessions")
        headers = {"Content-Type": "application/json"}
        if json_str != "":
            headers["Content-MD5"] = md5(to_binary(json_str)).hexdigest()
        params = {"session_type": "batch_read"}
        if self._quota_name:
            params["quotaName"] = self._quota_name

        res = self.tunnel_rest.post(url, data=json_str, params=params, headers=headers)

        response = TableBatchScanResponse()
        response.parse(res, obj=response)
        response.status = Status.OK if res.status_code == codes['created'] else Status.WAIT
        update_request_id(response, res)

        return response

    def get_read_session(self, request: SessionRequest) -> TableBatchScanResponse:
        """Get the read session.

        Args:
            request: Read session parameters sent to the server.

        Returns:
            Read session response returned from the server.
        """
        if not isinstance(request, SessionRequest):
            raise ValueError("Use SessionRequest class to build request for get read session interface")

        url = self._get_resource("sessions", request.session_id)
        headers = {}
        params = {"session_type": "batch_read"}
        if self._quota_name:
            params["quotaName"] = self._quota_name

        res = self.tunnel_rest.get(url, params=params, headers=headers)

        response = TableBatchScanResponse()
        response.parse(res, obj=response)
        response.status = Status.OK
        update_request_id(response, res)

        return response

    def read_rows_stream(self, request: ReadRowsRequest) -> StreamReader:
        """Read one split of the read session. Stream means the data read from server is serialized arrow record batch.

        Args:
            request: Batch split parameters sent to the server.

        Returns:
            Stream reader.
        """
        if not isinstance(request, ReadRowsRequest):
            raise ValueError("Use ReadRowsRequest class to build request for read rows interface")

        url = self._get_resource("data")
        headers = {
            "Connection": "Keep-Alive",
            "Accept-Encoding": request.compression.name if request.compression != Compression.UNCOMPRESSED else ""
        }
        params = {
            "session_id": request.session_id,
            "max_batch_rows": str(request.max_batch_rows),
            "split_index": str(request.split_index),
            "row_count": str(request.row_count),
            "row_index": str(request.row_index)
        }
        if self._quota_name:
            params["quotaName"] = self._quota_name
        if request.data_format.type is not None:
            params["data_format_type"] = request.data_format.type
        if request.data_format.version is not None:
            params["data_format_version"] = request.data_format.version

        def download():
            return self.tunnel_rest.get(url, stream=True, params=params, headers=headers)

        return StreamReader(download)

    def create_write_session(self, request: TableBatchWriteRequest) -> TableBatchWriteResponse:
        """Create a write session.

        Args:
            request: Table write parameters sent to the server.

        Returns:
            Write session response returned from the server.
        """
        if not isinstance(request, TableBatchWriteRequest):
            raise ValueError("Use TableBatchWriteRequest class to build request for create write session interface")

        json_str = request.serialize()

        url = self._get_resource("sessions")
        headers = {"Content-Type": "application/json"}
        if json_str != "":
            headers["Content-MD5"] = md5(to_binary(json_str)).hexdigest()
        params = {"session_type": "batch_write"}
        if self._quota_name:
            params["quotaName"] = self._quota_name

        res = self.tunnel_rest.post(url, data=json_str, params=params, headers=headers)

        response = TableBatchWriteResponse()
        response.parse(res, obj=response)
        response.status = Status.OK
        update_request_id(response, res)

        return response

    def get_write_session(self, request: SessionRequest) -> TableBatchWriteResponse:
        """Get a write session.

        Args:
            request: Write session parameters sent to the server.

        Returns:
            Write session response returned from the server.
        """
        if not isinstance(request, SessionRequest):
            raise ValueError("Use SessionRequest class to build request for get write session interface")

        url = self._get_resource("sessions", request.session_id)
        headers = {}
        params = {"session_type": "batch_write"}
        if self._quota_name:
            params["quotaName"] = self._quota_name

        res = self.tunnel_rest.get(url, params=params, headers=headers)

        response = TableBatchWriteResponse()
        response.parse(res, obj=response)
        response.status = Status.OK
        update_request_id(response, res)

        return response

    def write_rows_stream(self, request: WriteRowsRequest) -> StreamWriter:
        """Write one block of data to the write session. Stream means the data written to server is serialized arrow record batch.

        Args:
            request: Batch write parameters sent to the server.

        Returns:
            Stream writer.
        """
        if not isinstance(request, WriteRowsRequest):
            raise ValueError("Use WriteRowsRequest class to build request for write rows interface")

        url = self._get_resource("sessions", request.session_id, "data")
        headers = {
            "Content-Type": "application/octet-stream",
            "Transfer-Encoding": "chunked"
        }

        params = {
            "attempt_number": str(request.attempt_number),
            "block_number": str(request.block_number)
        }
        if self._quota_name:
            params["quotaName"] = self._quota_name
        if request.data_format.type != None:
            params["data_format_type"] = str(request.data_format.type)
        if request.data_format.version != None:
            params["data_format_version"] = str(request.data_format.version)

        def upload(data):
            return self.tunnel_rest.post(url, data=data, params=params, headers=headers)

        return StreamWriter(upload)

    def commit_write_session(self, request: SessionRequest, commit_msg: list) -> TableBatchWriteResponse:
        """Commit the write session after write the last stream data.

        Args:
            request: Commit write session parameters sent to the server.
            commit_msg: Commit messages collected from the write_rows_stream().

        Returns:
            Write session response returned from the server.
        """
        if not isinstance(request, SessionRequest):
            raise ValueError("Use SessionRequest class to build request for commit write session interface")
        if not isinstance(commit_msg, list):
            raise ValueError("Use list for commit message")

        commit_message_dict = {"CommitMessages": commit_msg}
        json_str = json.dumps(commit_message_dict)

        url = self._get_resource("commit")
        headers = {"Content-Type": "application/json"}
        params = {"session_id": request.session_id}
        if self._quota_name:
            params["quotaName"] = self._quota_name

        res = self.tunnel_rest.post(url, data=json_str, params=params, headers=headers)

        response = TableBatchWriteResponse()
        response.parse(res, obj=response)
        response.status = Status.OK if res.status_code == codes['created'] else Status.WAIT
        update_request_id(response, res)

        return response


class StorageApiArrowClient(StorageApiClient):
    """Arrow batch client to bundle configuration needed for API requests."""
    def read_rows_arrow(self, request: ReadRowsRequest) -> ArrowReader:
        """Read one split of the read session.

        Args:
            request: Arrow batch split parameters sent to the server.

        Returns:
            Arrow batch reader.
        """
        if not isinstance(request, ReadRowsRequest):
            raise ValueError("Use ReadRowsRequest class to build request for read rows interface")

        return ArrowReader(self.read_rows_stream(request))

    def write_rows_arrow(self, request: WriteRowsRequest) -> ArrowWriter:
        """Write one block of data to the write session.

        Args:
            request: Arrow batch write parameters sent to the server.

        Returns:
            Arrow batch writer.
        """
        if not isinstance(request, WriteRowsRequest):
            raise ValueError("Use WriteRowsRequest class to build request for write rows interface")

        return ArrowWriter(self.write_rows_stream(request), request.compression)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

from .core import DataFrame, CollectionExpr
from .expr.expressions import SequenceExpr, Scalar, BuiltinFunction, RandomScalar
from .expr.element import switch, FuncFactory
from .expr.datetimes import year, month, day, hour, minute, second, millisecond
from .expr.composites import make_list, make_dict
from .expr.reduction import aggregate, agg
from .utils import output_types, output_names, output
from .delay import Delay
from ..compat import six


def NullScalar(tp):
    """
    Creates a Scalar representing typed None values.

    :param tp: type of the scalar
    :return: Scalar with None value
    """
    return Scalar(_value_type=tp)


func = FuncFactory()

try:
    import pandas as pd
    from pandas.io.api import *

    def wrap(func):
        def inner(*args, **kwargs):
            as_type = kwargs.pop("as_type", None)
            unknown_as_string = kwargs.pop("unknown_as_string", False)

            res = func(*args, **kwargs)
            if isinstance(res, pd.DataFrame):
                return DataFrame(
                    res, as_type=as_type, unknown_as_string=unknown_as_string
                )
            return res

        return inner

    for k, v in six.iteritems(dict(locals())):
        if k.startswith("read_") and inspect.isfunction(v):
            locals()[k] = wrap(v)
except ImportError:
    pass

try:
    from ..internal import df as internal_df

    del internal_df
except ImportError:
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque, OrderedDict
import itertools

from ..expr.dynamic import DynamicMixin
from ..expr.expressions import \
    FilterCollectionExpr, FilterPartitionCollectionExpr, Column, Scalar
from ..expr.arithmetic import Or, And, Equal
from ..expr.errors import ExpressionError
from ..utils import is_source_partition
from ...errors import ODPSError
from ... import compat, options


def refresh_dynamic(executed, dag, func=None):
    q = deque()
    q.append(executed)

    while len(q) > 0:
        curr = q.popleft()

        for p in dag.successors(curr):
            if isinstance(p, DynamicMixin):
                q.append(p)

        if isinstance(curr, DynamicMixin):
            deps = set([id(d) for d in curr.deps or []])
            if any(id(c) not in deps and isinstance(c, DynamicMixin) for c in curr.children()):
                continue
            try:
                sub = curr.to_static()
            except ExpressionError:
                # the node is the optimized to disappear which should not be handled here
                continue
            dag.substitute(curr, sub)
            if func:
                func(sub)


def _fetch_partition_size(part, table):
    if len(part.split(',')) < len(table.table_schema.partitions):
        try:
            return sum(p.size for p in table.iterate_partitions(spec=part))
        except ODPSError:
            return
    else:
        try:
            if part not in table.partitions:
                return
        except ODPSError:
            return

        return table.partitions[part].size


def fetch_data_source_size(expr_dag, node, table):
    schema = table.table_schema

    if not schema.partitions:
        # not partitioned table
        return table.size

    size = 0
    for parent in expr_dag.successors(node):
        if isinstance(parent, FilterPartitionCollectionExpr):
            partition_predicates = parent._predicate_string.split('/')
            for p in partition_predicates:
                curr_size = _fetch_partition_size(p, table)
                if curr_size:
                    size += curr_size
                else:
                    return
        elif isinstance(parent, FilterCollectionExpr):
            def gen(n):
                if isinstance(n, Equal):
                    expr = n.lhs
                    if isinstance(expr, Column) and is_source_partition(expr, next(expr.data_source())) \
                            and isinstance(n.rhs, Scalar):
                        return '%s=%s' % (expr.source_name, n.rhs.value)
                elif isinstance(n, And):
                    left_partition_spec = gen(n.lhs)
                    right_partition_spec = gen(n.rhs)
                    if left_partition_spec and right_partition_spec:
                        partition_spec = ','.join((left_partition_spec, right_partition_spec))
                    elif left_partition_spec:
                        partition_spec = left_partition_spec
                    elif right_partition_spec:
                        partition_spec = right_partition_spec
                    else:
                        return
                    return partition_spec

            def walk(predicate):
                if isinstance(predicate, Or):
                    left_size = walk(predicate._lhs)
                    right_size = walk(predicate._rhs)
                    return (left_size or 0) + (right_size or 0)
                elif isinstance(predicate, (And, Equal)):
                    partition_spec = gen(predicate)
                    curr_size = _fetch_partition_size(partition_spec, table) \
                        if partition_spec is not None else None
                    if curr_size is not None:
                        return curr_size

            filter_size = walk(parent._predicate)
            if not filter_size:
                return
            size += filter_size

    return size


def _convert_pd_type(values, table):
    import pandas as pd

    retvals = []
    for val, t in compat.izip(values, table.table_schema.types):
        isnull = pd.isnull(val)
        if isinstance(isnull, bool) and isnull:
            retvals.append(None)
        else:
            retvals.append(val)

    return retvals


def _reorder_pd(frame, table, cast=False):
    import pandas as pd
    from .odpssql.types import df_schema_to_odps_schema, df_type_to_odps_type, odps_type_to_df_type

    from .pd.types import df_type_to_np_type
    from .errors import CompileError

    expr_table_schema = df_schema_to_odps_schema(frame.schema).to_ignorecase_schema()
    for col in expr_table_schema.columns:
        if col.name.lower() not in table.table_schema:
            raise CompileError('Column(%s) does not exist in target table %s, '
                               'writing cannot be performed.' % (col.name, table.name))
        t_col = table.table_schema[col.name.lower()]
        if not cast and not t_col.type.can_implicit_cast(col.type):
            raise CompileError('Cannot implicitly cast column %s from %s to %s.' % (
                col.name, col.type, t_col.type))

    size = len(frame.values)

    df_col_dict = dict((c.name.lower(), c) for c in frame.columns)
    case_dict = dict((c.name.lower(), c.name) for c in frame.columns)

    data_dict = dict()
    for dest_col in table.table_schema.columns:
        if dest_col.name not in df_col_dict:
            data_dict[dest_col.name] = [None] * size
        else:
            src_type = df_type_to_odps_type(
                df_col_dict[dest_col.name].type, project=table.project
            )
            if src_type == dest_col.type:
                data_dict[dest_col.name] = frame.values[case_dict[dest_col.name]]
            elif dest_col.type.can_implicit_cast(src_type) or cast:
                new_np_type = df_type_to_np_type(odps_type_to_df_type(dest_col.type))
                data_dict[dest_col.name] = frame.values[case_dict[dest_col.name]].astype(new_np_type)
            else:
                raise CompileError('Column %s\'s type does not match, expect %s, got %s' % (
                    dest_col.name, src_type, dest_col.type))
    return pd.DataFrame(data_dict, columns=[c.name for c in table.table_schema.columns])


def _get_cached_writer(writer_cache, table, partition=None):
    writer_key = (table.name, partition)
    if writer_key not in writer_cache:
        if len(writer_cache) >= options.df.writer_count_limit:
            writer = writer_cache.popitem(last=False)
            writer.close()

        writer = writer_cache[writer_key] = table.open_writer(
            partition=partition
        )
    else:
        # manually put to end
        writer = writer_cache.pop(writer_key)
        writer_cache[writer_key] = writer
    return writer


def _write_table_no_partitions(
    frame, table, ui, writer_cache, created_partitions, cast=False, overwrite=True,
    partition=None, progress_proportion=1,
):
    def gen():
        df = _reorder_pd(frame, table, cast=cast)
        size = len(df)

        last_percent = 0
        for i, row in zip(itertools.count(), df.values):
            if i % 50 == 0:
                percent = float(i) / size * progress_proportion
                ui.inc(percent - last_percent)
                last_percent = percent

            yield table.new_record(_convert_pd_type(row, table))

        if last_percent < progress_proportion:
            ui.inc(progress_proportion - last_percent)

    if partition not in created_partitions:
        created_partitions.add(partition)
        if overwrite:
            if partition is None:
                table.truncate()
            else:
                table.delete_partition(partition, if_exists=True)
                table.create_partition(partition)

    _get_cached_writer(writer_cache, table, partition).write(gen())


def _write_table_with_partitions(
    frame, table, partitions, ui, writer_cache, created_partitions, cast=False,
    overwrite=True, progress_proportion=1,
):
    df = _reorder_pd(frame, table, cast=cast)
    vals_to_partitions = dict()
    for ps in df[partitions].drop_duplicates().values:
        p = ','.join('='.join([str(n), str(v)]) for n, v in zip(partitions, ps))
        if p not in created_partitions:
            if overwrite:
                table.delete_partition(p, if_exists=True)
            table.create_partition(p, if_not_exists=True)
            created_partitions.add(p)
        vals_to_partitions[tuple(ps)] = p

    size = len(df)
    curr = [0]
    last_percent = [0]
    for name, group in df.groupby(partitions):
        name = name if isinstance(name, tuple) else (name, )
        group = group[[it for it in group.columns.tolist() if it not in partitions]]

        def gen():
            for i, row in zip(itertools.count(), group.values):
                curr[0] += i
                if curr[0] % 50 == 0:
                    percent = float(curr[0]) / size * progress_proportion
                    ui.inc(percent - last_percent[0])
                    last_percent[0] = percent

                yield table.new_record(_convert_pd_type(row, table))

        _get_cached_writer(writer_cache, table, vals_to_partitions[name]).write(gen())

    if last_percent[0] < progress_proportion:
        ui.inc(progress_proportion - last_percent[0])


def write_table(frame, table, ui, cast=False, overwrite=True, partitions=None,
                partition=None, progress_proportion=1):
    from ..backends.frame import ResultFrame

    ui.status('Try to upload to ODPS with tunnel...')
    writers_cache = OrderedDict()
    created_partitions = set()
    for start in range(0, len(frame), options.tunnel.write_row_batch_size):
        subframe = frame[start: start + options.tunnel.write_row_batch_size]
        if not isinstance(subframe, ResultFrame):
            subframe = ResultFrame(subframe, schema=frame.schema)
        batch_proportion = 1.0 * progress_proportion * len(subframe) / len(frame)
        if partitions is None:
            _write_table_no_partitions(
                subframe, table, ui, writers_cache, created_partitions, cast=cast,
                overwrite=overwrite, partition=partition, progress_proportion=batch_proportion,
            )
        else:
            _write_table_with_partitions(
                subframe, table, partitions, ui, writers_cache, created_partitions,
                cast=cast, overwrite=overwrite, progress_proportion=batch_proportion,
            )
    for writer in writers_cache.values():
        writer.close()


def process_persist_kwargs(kw):
    kw.pop('partition', None)
    kw.pop('partitions', None)
    kw.pop('create_table', None)
    kw.pop('drop_table', None)
    kw.pop('create_partition', None)
    kw.pop('drop_partition', None)

    return kw


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function

import bisect
import calendar
import codecs
import copy
import glob
import hmac
import logging
import math
import multiprocessing
import os
import random
import re
import shutil
import string
import struct
import sys
import threading
import time
import traceback
import types
import uuid
import warnings
import xml.dom.minidom
from base64 import b64encode
from datetime import datetime, date, timedelta
from email.utils import parsedate_tz, formatdate
from hashlib import sha1, md5

try:
    from collections.abc import Hashable, Mapping, Iterable
except ImportError:
    from collections import Hashable, Mapping, Iterable

from . import compat, options
from .compat import six, getargspec, FixedOffset, parsedate_to_datetime, utc

try:
    import zoneinfo
except ImportError:
    zoneinfo = None
try:
    import pytz
except ImportError:
    pytz = None

try:
    from odps.src.utils_c import CMillisecondsConverter
except ImportError:
    CMillisecondsConverter = None

TEMP_TABLE_PREFIX = "tmp_pyodps_"
if six.PY3:  # make flake8 happy
    unicode = str

_IS_WINDOWS = sys.platform.lower().startswith("win")

logger = logging.getLogger(__name__)

notset = object()


def deprecated(msg, cond=None):
    def _decorator(func):
        """This is a decorator which can be used to mark functions
        as deprecated. It will result in a warning being emmitted
        when the function is used."""
        @six.wraps(func)
        def _new_func(*args, **kwargs):
            warn_msg = "Call to deprecated function %s." % func.__name__
            if isinstance(msg, six.string_types):
                warn_msg += ' ' + msg
            if cond is None or cond():
                warnings.warn(warn_msg, category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return _new_func

    if isinstance(msg, (types.FunctionType, types.MethodType)):
        return _decorator(msg)
    return _decorator


class ExperimentalNotAllowed(Exception):
    pass


def experimental(msg, cond=None):
    warn_cache = set()

    def _decorator(func):
        @six.wraps(func)
        def _new_func(*args, **kwargs):
            warn_msg = "Call to experimental function %s." % func.__name__
            if isinstance(msg, six.string_types):
                warn_msg += ' ' + msg

            if not str_to_bool(os.environ.get('PYODPS_EXPERIMENTAL', 'true')):
                err_msg = "Calling to experimental method %s is denied." % func.__name__
                if isinstance(msg, six.string_types):
                    err_msg += ' ' + msg
                raise ExperimentalNotAllowed(err_msg)

            if func not in warn_cache and (cond is None or cond()):
                warnings.warn(warn_msg, category=FutureWarning, stacklevel=2)
                warn_cache.add(func)
            return func(*args, **kwargs)

        # intentionally eliminate __doc__ for Volume 2
        _new_func.__doc__ = None
        return _new_func

    if isinstance(msg, (types.FunctionType, types.MethodType)):
        return _decorator(msg)
    return _decorator


def fixed_writexml(self, writer, indent="", addindent="", newl=""):
    # indent = current indentation
    # addindent = indentation to add to higher levels
    # newl = newline string
    writer.write(indent + "<" + self.tagName)

    attrs = self._get_attributes()
    a_names = compat.lkeys(attrs)
    a_names.sort()

    for a_name in a_names:
        writer.write(" %s=\"" % a_name)
        xml.dom.minidom._write_data(writer, attrs[a_name].value)
        writer.write("\"")
    if self.childNodes:
        if len(self.childNodes) == 1 \
          and self.childNodes[0].nodeType == xml.dom.minidom.Node.TEXT_NODE:
            writer.write(">")
            self.childNodes[0].writexml(writer, "", "", "")
            writer.write("</%s>%s" % (self.tagName, newl))
            return
        writer.write(">%s"%(newl))
        for node in self.childNodes:
            node.writexml(writer,indent+addindent,addindent,newl)
        writer.write("%s</%s>%s" % (indent,self.tagName,newl))
    else:
        writer.write("/>%s"%(newl))
# replace minidom's function with ours
xml.dom.minidom.Element.writexml = fixed_writexml
xml_fixed = lambda: None


def hmac_sha1(secret, data):
    return b64encode(hmac.new(secret, data, sha1).digest())


def md5_hexdigest(data):
    return md5(data).hexdigest()


def rshift(val, n):
    return val >> n if val >= 0  else \
        (val+0x100000000) >> n


def long_bits_to_double(bits):
    """
    @type  bits: long
    @param bits: the bit pattern in IEEE 754 layout

    @rtype:  float
    @return: the double-precision floating-point value corresponding
             to the given bit pattern C{bits}.
    """
    return struct.unpack('d', struct.pack('Q', bits))[0]


def double_to_raw_long_bits(value):
    """
    @type  value: float
    @param value: a Python (double-precision) float value

    @rtype: long
    @return: the IEEE 754 bit representation (64 bits as a long integer)
             of the given double-precision floating-point value.
    """
    # pack double into 64 bits, then unpack as long int
    return struct.unpack('Q', struct.pack('d', float(value)))[0]


def camel_to_underline(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def underline_to_capitalized(name):
    return "".join([s[0].upper() + s[1:len(s)] for s in name.strip('_').split('_')])


def underline_to_camel(name):
    parts = name.split('_')
    return parts[0] + ''.join(v.title() for v in parts[1:])


def long_to_int(value):
    if value & 0x80000000:
        return int(-((value ^ 0xFFFFFFFF) + 1))
    else:
        return int(value)


def int_to_uint(v):
    if v < 0:
        return int(v + 2**32)
    return v


def long_to_uint(value):
    v = long_to_int(value)
    return int_to_uint(v)


def stringify_expt():
    lines = traceback.format_exception(*sys.exc_info())
    return '\n'.join(lines)


def str_to_printable(field_name, auto_quote=True):
    if not field_name:
        return field_name

    escapes = {'\\': '\\\\', '\'': '\\\'', '"': '\\"', '\a': '\\a', '\b': '\\b', '\f': '\\f',
               '\n': '\\n', '\r': '\\r', '\t': '\\t', '\v': '\\v', ' ': ' '}

    def _escape_char(c):
        if c in escapes:
            return escapes[c]
        elif c < ' ':
            return '\\x%02x' % ord(c)
        else:
            return c

    need_escape = lambda c: c <= ' ' or c in escapes
    if any(need_escape(c) for c in field_name):
        ret = ''.join(_escape_char(ch) for ch in field_name)
        return '"' + ret + '"' if auto_quote else ret
    return field_name


def indent(text, n_spaces):
    if n_spaces <= 0:
        return text
    block = ' ' * n_spaces
    return '\n'.join((block + it) if len(it) > 0 else it
                     for it in text.split('\n'))


def parse_rfc822(s, use_legacy_parsedate=None):
    if s is None:
        return None

    use_legacy_parsedate = (
        use_legacy_parsedate if use_legacy_parsedate is not None
        else options.use_legacy_parsedate
    )
    if use_legacy_parsedate:
        date_tuple = parsedate_tz(s)
        return datetime(*date_tuple[:6])

    time_obj = parsedate_to_datetime(s)
    if time_obj.tzinfo is None:
        return time_obj
    time_obj = time_obj.astimezone(utc)
    gmt_ts = calendar.timegm(time_obj.timetuple())
    return datetime.fromtimestamp(gmt_ts)


def gen_rfc822(dt=None, localtime=False, usegmt=False):
    if dt is not None:
        t = time.mktime(dt.timetuple())
    else:
        t = None
    return formatdate(t, localtime=localtime, usegmt=usegmt)


try:
    _antique_mills = time.mktime(datetime(1928, 1, 1).timetuple()) * 1000
except OverflowError:
    _antique_mills = int(
        (datetime(1928, 1, 1) - datetime.utcfromtimestamp(0)).total_seconds()
    ) * 1000
_min_datetime_mills = int(
    (datetime.min - datetime.utcfromtimestamp(0)).total_seconds() * 1000
)
_antique_errmsg = 'Date older than 1928-01-01 and may contain errors. ' \
                  'Ignore this error by configuring `options.allow_antique_date` to True.'
_min_datetime_errmsg = 'Date exceed range Python can handle. If you are reading data with tunnel, read '\
                       'the value as None by setting options.tunnel.overflow_date_as_none to True, ' \
                       'or convert the value into strings with SQL before processing them with Python.'


def to_timestamp(dt, local_tz=None, is_dst=False):
    return int(to_milliseconds(dt, local_tz=local_tz, is_dst=is_dst) / 1000.0)


class MillisecondsConverter(object):
    _inst_cache = dict()

    @classmethod
    def _get_tz(cls, tz):
        if isinstance(tz, six.string_types):
            if pytz is None and zoneinfo is None:
                raise ImportError('Package `pytz` is needed when specifying string-format time zone.')
            else:
                return get_zone_from_name(tz)
        else:
            return tz

    def __new__(cls, local_tz=None, is_dst=False):
        cache_key = (cls, local_tz, is_dst)
        if cache_key in cls._inst_cache:
            return cls._inst_cache[cache_key]
        o = super(MillisecondsConverter, cls).__new__(cls)
        o.__init__(local_tz, is_dst)
        cls._inst_cache[cache_key] = o
        return o

    def _windows_mktime(self, timetuple):
        if self._local_tz:
            fromtimestamp = datetime.fromtimestamp
            mktime = time.mktime
        else:
            fromtimestamp = datetime.utcfromtimestamp
            mktime = calendar.timegm

        if timetuple[0] > 1970:
            return mktime(timetuple)
        dt = datetime(*timetuple[:6])
        epoch = fromtimestamp(0)
        return int((dt - epoch).total_seconds())

    def _windows_fromtimestamp(self, seconds):
        fromtimestamp = datetime.fromtimestamp if self._local_tz else datetime.utcfromtimestamp
        if seconds >= 0:
            return fromtimestamp(seconds)
        epoch = fromtimestamp(0)
        return epoch + timedelta(seconds=seconds)

    def __init__(self, local_tz=None, is_dst=False):
        self._local_tz = local_tz if local_tz is not None else options.local_timezone
        if self._local_tz is None:
            self._local_tz = True
        self._use_default_tz = type(self._local_tz) is bool

        self._allow_antique = options.allow_antique_date or _antique_mills is None
        self._is_dst = is_dst

        if self._local_tz:
            self._mktime = time.mktime
            self._fromtimestamp = datetime.fromtimestamp
        else:
            self._mktime = calendar.timegm
            self._fromtimestamp = datetime.utcfromtimestamp

        if _IS_WINDOWS:
            # special logic for negative timestamp under Windows
            self._mktime = self._windows_mktime
            self._fromtimestamp = self._windows_fromtimestamp

        self._tz = self._get_tz(self._local_tz) if not self._use_default_tz else None
        if hasattr(self._tz, 'localize'):
            self._localize = lambda dt: self._tz.localize(dt, is_dst=is_dst)
        else:
            self._localize = lambda dt: dt.replace(tzinfo=self._tz)

    def to_milliseconds(self, dt):
        from .errors import DatetimeOverflowError

        if not self._use_default_tz and dt.tzinfo is None:
            dt = self._localize(dt)

        if dt.tzinfo is not None:
            mills = int((calendar.timegm(
                dt.astimezone(compat.utc).timetuple()) + dt.microsecond / 1000000.0
            ) * 1000)
        else:
            mills = int((self._mktime(dt.timetuple()) + dt.microsecond / 1000000.0) * 1000)

        if not self._allow_antique and mills < _antique_mills:
            raise DatetimeOverflowError(_antique_errmsg)
        return mills

    def from_milliseconds(self, milliseconds):
        from .errors import DatetimeOverflowError

        if not self._allow_antique and milliseconds < _antique_mills:
            raise DatetimeOverflowError(_antique_errmsg)
        if milliseconds < _min_datetime_mills:
            raise DatetimeOverflowError(_min_datetime_errmsg)

        seconds = compat.long_type(math.floor(milliseconds / 1000))
        microseconds = compat.long_type(milliseconds) % 1000 * 1000
        if self._use_default_tz:
            return self._fromtimestamp(seconds).replace(microsecond=microseconds)
        else:
            return datetime.utcfromtimestamp(seconds)\
                .replace(microsecond=microseconds, tzinfo=compat.utc)\
                .astimezone(self._tz)


def to_milliseconds(dt, local_tz=None, is_dst=False, force_py=False):
    cls = CMillisecondsConverter
    if force_py or cls is None:
        cls = MillisecondsConverter
    f = cls(local_tz, is_dst=is_dst)
    return f.to_milliseconds(dt)


def to_days(dt):
    start_day = date(1970, 1, 1)
    return (dt - start_day).days


def to_date(delta_day):
    start_day = date(1970, 1, 1)
    return start_day + timedelta(delta_day)


def to_datetime(milliseconds, local_tz=None, force_py=False):
    cls = CMillisecondsConverter
    if force_py or cls is None:
        cls = MillisecondsConverter
    f = cls(local_tz)
    return f.from_milliseconds(milliseconds)


def strptime_with_tz(dt, format='%Y-%m-%d %H:%M:%S'):
    try:
        return datetime.strptime(dt, format)
    except ValueError:
        naive_date_str, _, offset_str = dt.rpartition(' ')
        naive_dt = datetime.strptime(naive_date_str, format)
        offset = int(offset_str[-4:-2]) * 60 + int(offset_str[-2:])
        if offset_str[0] == "-":
            offset = -offset
        return naive_dt.replace(tzinfo=FixedOffset(offset))


def to_binary(text, encoding='utf-8'):
    if text is None:
        return text
    if isinstance(text, six.text_type):
        return text.encode(encoding)
    elif isinstance(text, (six.binary_type, bytearray)):
        return bytes(text)
    else:
        return str(text).encode(encoding) if six.PY3 else str(text)


def to_text(binary, encoding='utf-8'):
    if binary is None:
        return binary
    if isinstance(binary, (six.binary_type, bytearray)):
        return binary.decode(encoding)
    elif isinstance(binary, six.text_type):
        return binary
    else:
        return str(binary) if six.PY3 else str(binary).decode(encoding)


def to_str(text, encoding='utf-8'):
    return to_text(text, encoding=encoding) if six.PY3 else to_binary(text, encoding=encoding)


def get_zone_from_name(tzname):
    return zoneinfo.ZoneInfo(tzname) if zoneinfo else pytz.timezone(tzname)


def get_zone_name(tz):
    return getattr(tz, "key", None) or getattr(tz, "zone", None)


# fix encoding conversion problem under windows
if sys.platform == 'win32':
    def _replace_default_encoding(func):
        def _fun(s, encoding=None):
            return func(s, encoding=encoding or options.display.encoding)

        _fun.__name__ = func.__name__
        _fun.__doc__ = func.__doc__
        return _fun

    to_binary = _replace_default_encoding(to_binary)
    to_text = _replace_default_encoding(to_text)
    to_str = _replace_default_encoding(to_str)


def is_lambda(f):
    lam = lambda: 0
    return isinstance(f, type(lam)) and f.__name__ == lam.__name__


def str_to_kv(string, typ=None):
    d = dict()
    for pair in string.split(','):
        k, v = pair.split(':', 1)
        if typ:
            v = typ(v)
        d[k] = v
    return d


def interval_select(val, intervals, targets):
    return targets[bisect.bisect_left(intervals, val)]


def is_namedtuple(obj):
    return isinstance(obj, tuple) and hasattr(obj, '_fields')


def str_to_bool(s):
    if isinstance(s, bool):
        return s
    s = s.lower().strip()
    if s == 'true':
        return True
    elif s == 'false':
        return False
    else:
        raise ValueError


def bool_to_str(s):
    return str(s).lower()


def get_root_dir():
    return os.path.dirname(sys.modules[__name__].__file__)


def load_text_file(path):
    file_path = get_root_dir() + path
    if not os.path.exists(file_path):
        return None
    with codecs.open(file_path, encoding='utf-8') as f:
        inp_file = f.read()
        f.close()
    return inp_file


def load_file_paths(pattern):
    file_path = os.path.normpath(os.path.dirname(sys.modules[__name__].__file__) + pattern)
    return glob.glob(file_path)


def load_static_file_paths(path):
    return load_file_paths('/static/' + path)


def load_text_files(pattern, func=None):
    file_path = os.path.normpath(os.path.dirname(sys.modules[__name__].__file__) + pattern)
    content_dict = dict()
    for file_path in glob.glob(file_path):
        _, fn = os.path.split(file_path)
        if func and not func(fn):
            continue
        with codecs.open(file_path, encoding='utf-8') as f:
            content_dict[fn] = f.read()
            f.close()
    return content_dict


def load_static_text_file(path):
    return load_text_file('/static/' + path)


def load_internal_static_text_file(path):
    return load_text_file('/internal/static/' + path)


def load_static_text_files(pattern, func=None):
    return load_text_files('/static/' + pattern, func)


def init_progress_bar(val=1, use_console=True):
    try:
        from traitlets import TraitError
        ipython = True
    except ImportError:
        try:
            from IPython.utils.traitlets import TraitError
            ipython = True
        except ImportError:
            ipython = False

    from .console import ProgressBar, is_widgets_available

    if not ipython:
        bar = ProgressBar(val) if use_console else None
    else:
        try:
            if is_widgets_available():
                bar = ProgressBar(val, True)
            else:
                bar = ProgressBar(val) if use_console else None
        except TraitError:
            bar = ProgressBar(val) if use_console else None

    return bar


def init_progress_ui(val=1, lock=False, use_console=True, mock=False):
    from .ui import ProgressGroupUI, html_notify

    progress_group = None
    bar = None
    if not mock and is_main_thread():
        bar = init_progress_bar(val=val, use_console=use_console)
        if bar and bar._ipython_widget:
            try:
                progress_group = ProgressGroupUI(bar._ipython_widget)
            except:
                pass

    _lock = threading.Lock() if lock else None

    def ui_method(func):
        def inner(*args, **kwargs):
            if mock:
                return
            if _lock:
                with _lock:
                    return func(*args, **kwargs)
            else:
                return func(*args, **kwargs)
        return inner

    class ProgressUI(object):
        @ui_method
        def update(self, value=None):
            if bar:
                bar.update(value=value)

        @ui_method
        def current_progress(self):
            if bar and hasattr(bar, '_current_value'):
                return bar._current_value

        @ui_method
        def inc(self, value):
            if bar and hasattr(bar, '_current_value'):
                current_val = bar._current_value
                bar.update(current_val + value)

        @ui_method
        def status(self, prefix, suffix='', clear_keys=False):
            if progress_group:
                if clear_keys:
                    progress_group.clear_keys()
                progress_group.prefix = prefix
                progress_group.suffix = suffix

        @ui_method
        def add_keys(self, keys):
            if progress_group:
                progress_group.add_keys(keys)

        @ui_method
        def remove_keys(self, keys):
            if progress_group:
                progress_group.remove_keys(keys)

        @ui_method
        def update_group(self):
            if progress_group:
                progress_group.update()

        @ui_method
        def notify(self, msg):
            html_notify(msg)

        @ui_method
        def close(self):
            if bar:
                bar.close()
            if progress_group:
                progress_group.close()

    return ProgressUI()


def escape_odps_string(src):
    trans_dict = {
        "\b": r"\b",
        "\t": r"\t",
        "\n": r"\n",
        "\r": r"\r",
        "'": r"\'",
        '"': r"\"",
        "\\": r"\\",
        ";": r"\;",
        "Z": r"\Z",
        "\0": r"\0",
    }
    return "".join(trans_dict[ch] if ch in trans_dict else ch for ch in src)


def replace_sql_parameters(sql, ns):
    param_re = re.compile(r':([a-zA-Z_][a-zA-Z0-9_]*)')

    def is_numeric(val):
        return isinstance(val, (six.integer_types, float))

    def is_sequence(val):
        return isinstance(val, (tuple, set, list))

    def format_string(val):
        return "'{0}'".format(escape_odps_string(str(val)))

    def format_numeric(val):
        return repr(val)

    def format_sequence(val):
        escaped = [format_numeric(v) if is_numeric(v) else format_string(v) for v in val]
        return '({0})'.format(', '.join(escaped))

    def replace(matched):
        name = matched.group(1)
        val = ns.get(name)
        if val is None:
            return matched.group(0)
        elif is_numeric(val):
            return format_numeric(val)
        elif is_sequence(val):
            return format_sequence(val)
        else:
            return format_string(val)

    return param_re.sub(replace, sql)


def is_main_process():
    return 'main' in multiprocessing.current_process().name.lower()


survey_calls = dict()


def survey(func):
    @six.wraps(func)
    def wrapped(*args, **kwargs):
        arg_spec = getargspec(func)

        if 'self' in arg_spec.args:
            func_cls = args[0].__class__
        else:
            func_cls = None

        if func_cls:
            func_sig = '.'.join([func_cls.__module__, func_cls.__name__, func.__name__])
        else:
            func_sig = '.'.join([func.__module__, func.__name__])

        add_survey_call(func_sig)
        return func(*args, **kwargs)

    return wrapped


def add_survey_call(group):
    if any(r.search(group) is not None for r in options.skipped_survey_regexes):
        return
    if group not in survey_calls:
        survey_calls[group] = 1
    else:
        survey_calls[group] += 1


def get_survey_calls():
    return copy.copy(survey_calls)


def clear_survey_calls():
    survey_calls.clear()


def require_package(pack_name):
    def _decorator(func):
        try:
            __import__(pack_name, fromlist=[''])
            return func
        except ImportError:
            return None

    return _decorator


def gen_repr_object(**kwargs):
    obj = type('ReprObject', (), {})
    text = kwargs.pop('text', None)
    if six.PY2 and isinstance(text, unicode):
        text = text.encode('utf-8')
    if text:
        setattr(obj, 'text', text)
        setattr(obj, '__repr__', lambda self: text)
    for k, v in six.iteritems(kwargs):
        setattr(obj, k, v)
        setattr(obj, '_repr_{0}_'.format(k), lambda self: v)
    if 'gv' in kwargs:
        try:
            from graphviz import Source
            setattr(obj, '_repr_svg_', lambda self: Source(self._repr_gv_(), encoding='utf-8')._repr_svg_())
        except ImportError:
            pass
    return obj()


def build_pyodps_dir(*args):
    default_dir = os.path.join(os.path.expanduser('~'), '.pyodps')
    if sys.platform == 'win32' and 'APPDATA' in os.environ:
        win_default_dir = os.path.join(os.environ['APPDATA'], 'pyodps')
        if os.path.exists(default_dir):
            shutil.move(default_dir, win_default_dir)
        default_dir = win_default_dir
    home_dir = os.environ.get('PYODPS_DIR') or default_dir
    return os.path.join(home_dir, *args)


def object_getattr(obj, attr, default=None):
    try:
        return object.__getattribute__(obj, attr)
    except AttributeError:
        return default


def attach_internal(cls):
    cls_path = cls.__module__ + '.' + cls.__name__
    try:
        from .internal.core import MIXIN_TARGETS
        mixin_cls = MIXIN_TARGETS[cls_path]
        for method_name in dir(mixin_cls):
            if method_name.startswith('_'):
                continue
            att = getattr(mixin_cls, method_name)
            if six.PY2 and type(att).__name__ in ('instancemethod', 'method'):
                att = att.__func__
            setattr(cls, method_name, att)
        return cls
    except ImportError:
        return cls


def is_main_thread():
    if hasattr(threading, "main_thread"):
        return threading.current_thread() is threading.main_thread()
    return threading.current_thread().__class__.__name__ == '_MainThread'


def write_log(msg):
    """Legacy method to keep compatibility"""
    logger.info(msg)


def split_quoted(s, delimiter=',', maxsplit=0):
    pattern = r"""((?:[^""" + delimiter + r""""']|"[^"]*"|'[^']*')+)"""
    return re.split(pattern, s, maxsplit=maxsplit)[1::2]


def gen_temp_table():
    return '%s%s' % (TEMP_TABLE_PREFIX, str(uuid.uuid4()).replace('-', '_'))


def hashable(obj):
    if isinstance(obj, Hashable):
        items = obj
    elif isinstance(obj, Mapping):
        items = type(obj)((k, hashable(v)) for k, v in six.iteritems(obj))
    elif isinstance(obj, Iterable):
        items = tuple(hashable(item) for item in obj)
    else:
        raise TypeError(type(obj))

    return items


def thread_local_attribute(thread_local_name, default_value=None):
    attr_name = '_local_attr_%d' % random.randint(0, 99999999)

    def _get_thread_local(self):
        thread_local = getattr(self, thread_local_name, None)
        if thread_local is None:
            setattr(self, thread_local_name, threading.local())
            thread_local = getattr(self, thread_local_name)
        return thread_local

    def _getter(self):
        thread_local = _get_thread_local(self)
        if not hasattr(thread_local, attr_name) and callable(default_value):
            setattr(thread_local, attr_name, default_value())
        return getattr(thread_local, attr_name)

    def _setter(self, value):
        thread_local = _get_thread_local(self)
        setattr(thread_local, attr_name, value)

    return property(fget=_getter, fset=_setter)


def call_with_retry(func, *args, **kwargs):
    retry_num = 0
    retry_times = kwargs.pop("retry_times", options.retry_times)
    delay = kwargs.pop("delay", options.retry_delay)
    exc_type = kwargs.pop("exc_type", BaseException)
    while True:
        try:
            return func(*args, **kwargs)
        except exc_type:
            retry_num += 1
            time.sleep(delay)
            if retry_num > retry_times:
                raise


def get_id(n):
    if hasattr(n, '_node_id'):
        return n._node_id

    return id(n)


def strip_if_str(s):
    if isinstance(s, six.binary_type):
        s = to_str(s)
    if isinstance(s, six.string_types):
        return s.strip()
    return s


def with_wait_argument(func):
    func_spec = compat.getfullargspec(func) if compat.getfullargspec else compat.getargspec(func)
    args_set = set(func_spec.args)
    if hasattr(func_spec, "kwonlyargs"):
        args_set |= set(func_spec.kwonlyargs or [])
    has_varkw = (getattr(func_spec, "varkw", None) or getattr(func_spec, "keywords", None)) is not None

    try:
        async_index = func_spec.args.index("async_")
    except ValueError:
        async_index = None

    @six.wraps(func)
    def wrapped(*args, **kwargs):
        if async_index is not None and len(args) >= async_index + 1:
            warnings.warn(
                "Please use async_ as a keyword argument, like obj.func(async_=True)",
                DeprecationWarning, stacklevel=2
            )
            add_survey_call(".".join([func.__module__, func.__name__, "async_"]))
        elif "wait" in kwargs:
            kwargs["async_"] = not kwargs.pop("wait")
        elif "async" in kwargs:
            kwargs["async_"] = kwargs.pop("async")

        if not has_varkw and kwargs:
            no_args_match = [key for key in kwargs if key not in args_set]
            if no_args_match:
                warnings.warn(
                    "Arguments %s not supported, ignored by default. "
                    "Please check argument spellings." % (", ".join(no_args_match)),
                    stacklevel=2,
                )
            for arg in no_args_match:
                kwargs.pop(arg, None)
        return func(*args, **kwargs)

    return wrapped


def split_sql_by_semicolon(sql_statement):
    sql_statement = sql_statement.replace("\r\n", "\n").replace("\r", "\n")
    left_brackets = {'}': '{', ']': '[', ')': '('}

    def cut_statement(stmt_start, stmt_end=None):
        stmt_end = stmt_end or len(sql_statement)
        parts = []
        left = stmt_start
        for comm_start, comm_end in comment_blocks:
            if comm_end <= stmt_start:
                continue
            if comm_start > stmt_end:
                break
            parts.append(sql_statement[left: comm_start])
            left = comm_end
        parts.append(sql_statement[left: stmt_end])
        combined_lines = "".join(parts).splitlines()
        return "\n".join(line.rstrip() for line in combined_lines).strip()

    start, pos = 0, 0
    statements = []
    comment_sign = None
    comment_pos = None
    comment_blocks = []
    quote_sign = None
    bracket_stack = []
    while pos < len(sql_statement):
        ch = sql_statement[pos]
        dch = sql_statement[pos: pos + 2] if pos + 1 < len(sql_statement) else None
        if quote_sign is None and comment_sign is None:
            if ch in ('{', '[', '('):
                # start of brackets
                bracket_stack.append(ch)
                pos += 1
            elif ch in ('}', ']', ')'):
                # end of brackets
                assert bracket_stack[-1] == left_brackets[ch]
                bracket_stack.pop()
                pos += 1
            elif ch in ('"', "'", '`'):
                # start of quote
                quote_sign = ch
                pos += 1
            elif dch in ("--", "/*"):
                # start of line or block comments
                comment_sign = dch
                comment_pos = pos
                pos += 2
            elif ch == ';' and not bracket_stack:
                # semicolon without brackets, quotes and comments
                part_statement = cut_statement(start, pos + 1)
                if part_statement and part_statement != ';':
                    statements.append(part_statement)
                pos += 1
                start = pos
            else:
                pos += 1
        elif quote_sign is not None and ch == quote_sign:
            quote_sign = None
            pos += 1
        elif quote_sign is not None and ch == '\\':
            # skip escape char
            pos += 2
        elif comment_sign == "--" and ch == "\n":
            # line comment ends
            comment_sign = None
            comment_blocks.append((comment_pos, pos))
            pos += 1
        elif comment_sign == "/*" and dch == "*/":
            # block comment ends
            comment_sign = None
            comment_blocks.append((comment_pos, pos + 2))
            pos += 2
        else:
            pos += 1
    part_statement = cut_statement(start)
    if part_statement and part_statement != ';':
        statements.append(part_statement)
    return statements


_convert_host_hash = ("6fe9b6c02efc24159c09863c1aadffb5", "9b75728355160c10b5eb75e4bf105a76")
_default_host_hash = "c7f4116fbf820f99284dcc89f340b372"


def get_default_logview_endpoint(default_endpoint, odps_endpoint):
    try:
        if odps_endpoint is None:
            return default_endpoint

        parsed_host = compat.urlparse(odps_endpoint).hostname
        if parsed_host is None:
            return default_endpoint

        default_host = compat.urlparse(default_endpoint).hostname
        hashed_host = md5_hexdigest(to_binary(parsed_host))
        hashed_default_host = md5_hexdigest(to_binary(default_host))

        if hashed_default_host == _default_host_hash and hashed_host in _convert_host_hash:
            suffix = r"\1" + string.ascii_letters[1::-1] * 2 + parsed_host[-8:]
            return re.sub(r"([a-z])[a-z]{3}\.[a-z]{3}$", suffix, default_endpoint)
        return default_endpoint
    except:
        return default_endpoint


def _import_version_function():
    try:
        try:
            from importlib.metadata import version
        except ImportError:
            from importlib_metadata import version
        return version
    except ImportError:
        try:
            import pkg_resources
        except ImportError:
            return None
        return lambda x: pkg_resources.get_distribution(x).version


def get_package_version(pack_name, import_name=None):
    version_func = _import_version_function()
    if version_func is not None:
        try:
            return version_func(pack_name)
        except:
            pass
    import_name = pack_name if import_name is True else import_name
    if import_name:
        mod = __import__(import_name)
        return mod.__version__
    return None


def show_versions():  # pragma: no cover
    import locale
    import platform

    uname_result = platform.uname()
    language_code, encoding = locale.getlocale()

    results = {
        "python": ".".join([str(i) for i in sys.version_info]),
        "python-bits": struct.calcsize("P") * 8,
        "OS": uname_result.system,
        "OS-release": uname_result.release,
        "Version": uname_result.version,
        "machine": uname_result.machine,
        "processor": uname_result.processor,
        "byteorder": sys.byteorder,
        "LC_ALL": os.environ.get("LC_ALL"),
        "LANG": os.environ.get("LANG"),
        "LOCALE": {"language-code": language_code, "encoding": encoding},
    }

    try:
        from .src import crc32c_c  # noqa: F401

        results["USE_CLIB"] = True
    except ImportError:
        results["USE_CLIB"] = False

    try:
        from . import internal

        results["HAS_INTERNAL"] = True
    except ImportError:
        results["HAS_INTERNAL"] = False

    packages = {
        "pyodps": "odps",
        "requests": "requests",
        "urllib3": "urllib3",
        "charset_normalizer": "charset_normalizer",
        "chardet": "chardet",
        "idna": "idna",
        "certifi": "certifi",
        "numpy": "numpy",
        "scipy": "scipy",
        "pandas": "pandas",
        "matplotlib": "matplotlib",
        "sqlalchemy": "sqlalchemy",
        "pytz": "pytz",
        "python-dateutil": "dateutil",
        "IPython": "IPython",
        "maxframe": "maxframe",
    }
    for pack_name, pack_imp in packages.items():
        try:
            results[pack_name] = get_package_version(pack_name, pack_imp)
        except (ImportError, AttributeError):
            pass
    key_size = 1 + max(len(key) for key in results.keys())
    for key, val in results.items():
        print(key + " " * (key_size - len(key)) + ": " + str(val))



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from copy import copy
from functools import partial

from .classification import *
from .regression import *
from ...compat import six


class BaseScorer(six.with_metaclass(ABCMeta, object)):
    def __init__(self, func, sign, kwargs):
        self._func = func
        self._sign = sign
        self._kwargs = kwargs

    @abstractmethod
    def __call__(self, df, col_true=None, col_pred=None, sample_weight=None):
        pass


class PredictScorer(BaseScorer):
    def __call__(self, df, col_true=None, col_pred=None, sample_weight=None, **kwargs):
        kw = copy(self._kwargs)
        kw.update(kwargs)
        if col_true:
            kw['col_true'] = col_true
        if col_pred:
            kw['col_pred'] = col_pred
        if kw.get('execute_now', True):
            if sample_weight is not None:
                return self._sign * self._func(df, sample_weight=sample_weight, **kw)
            else:
                return self._sign * self._func(df, **kw)
        else:
            if sample_weight is not None:
                expr = self._func(df, sample_weight=sample_weight, **kw)
            else:
                expr = self._func(df, **kw)
            old_callback = getattr(expr, '_result_callback', None) or (lambda v: v)
            expr._result_callback = lambda v: self._sign * old_callback(v)
            return expr


def make_scorer(func, greater_is_better=True, **kwargs):
    sign = 1 if greater_is_better else -1
    return PredictScorer(func, sign, kwargs)

mean_squared_error_scorer = make_scorer(mean_squared_error,
                                        greater_is_better=False)
mean_absolute_error_scorer = make_scorer(mean_absolute_error,
                                         greater_is_better=False)
mean_absolute_percentage_scorer = make_scorer(mean_absolute_percentage_error,
                                              greater_is_better=False)

accuracy_scorer = make_scorer(accuracy_score)
roc_auc_scorer = make_scorer(roc_auc_score)

SCORERS = dict(mean_absolute_error=mean_absolute_error_scorer,
               mean_squared_error=mean_squared_error_scorer,
               mean_absolute_percentage_error=mean_absolute_percentage_scorer,
               accuracy=accuracy_scorer,
               roc_auc=roc_auc_scorer)

for name, metric in [('precision', precision_score),
                     ('recall', recall_score), ('f1', f1_score)]:
    SCORERS[name] = make_scorer(metric)
    for average in ['macro', 'micro', 'samples', 'weighted']:
        qualified_name = '{0}_{1}'.format(name, average)
        SCORERS[qualified_name] = make_scorer(partial(metric, pos_label=None,
                                                      average=average))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from datetime import datetime as _datetime, date as _date
from decimal import Decimal as _Decimal

from ..compat import six
from ..lib.xnamedtuple import xnamedtuple
from ..models import TableSchema, Column
from ..config import options
from ..types import DataType, Array, Map, Struct as _Struct, parse_composite_types


class Primitive(DataType):
    __slots__ = ()

    @property
    def CLASS_NAME(self):
        return self.__class__.__name__

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)

        return value


class Integer(Primitive):
    __slots__ = ()

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, Integer) and other._n_bytes <= self._n_bytes:
            return True
        return False

    def validate_value(self, val, max_field_size=None):
        if val is None and self.nullable:
            return True
        return self._bounds[0] <= val <= self._bounds[1]


class Float(Primitive):
    __slots__ = ()

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (Integer, Float)):
            return True
        return False

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)

        return float(value)


class Int8(Integer):
    __slots__ = ()

    _n_bytes = 1
    _bounds = (-128, 127)


class Int16(Integer):
    __slots__ = ()

    _n_bytes = 2
    _bounds = (-32768, 32767)


class Int32(Integer):
    __slots__ = ()

    _n_bytes = 4
    _bounds = (-2147483648, 2147483647)


class Int64(Integer):
    __slots__ = ()

    _n_bytes = 8
    _bounds = (-9223372036854775808, 9223372036854775807)


class Float32(Float):
    __slots__ = ()

    _n_bytes = 4


class Float64(Float):
    __slots__ = ()

    _n_bytes = 8


class Datetime(Primitive):
    __slots__ = ()

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (Datetime, String, Integer)):
            return True
        return False


class Date(Primitive):
    __slots__ = ()

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (Timestamp, Datetime, String)):
            return True
        return False


class Timestamp(Primitive):
    __slots__ = ()

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (Timestamp, Datetime, String)):
            return True
        return False


class Json(Primitive):
    __slots__ = ()


class Boolean(Primitive):
    __slots__ = ()


class Decimal(Primitive):
    __slots__ = ()

    def can_implicit_cast(self, other):
        if isinstance(other, (Decimal, Integer)):
            return True
        return False


class String(Primitive):
    __slots__ = ()


class Binary(Primitive):
    __slots__ = ()


class List(Array):
    __slots__ = ()
    CLASS_NAME = "List"

    def __init__(self, value_type, nullable=True):
        DataType.__init__(self, nullable=nullable)
        self.value_type = validate_data_type(value_type)

    def _equals(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return DataType._equals(self, other) and \
            self.value_type == other.value_type

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return isinstance(other, List) and \
            self.value_type == other.value_type and \
            self.nullable == other.nullable


class Dict(Map):
    __slots__ = ()
    CLASS_NAME = "Dict"

    def __init__(self, key_type, value_type, nullable=True):
        DataType.__init__(self, nullable=nullable)
        self.key_type = validate_data_type(key_type)
        self.value_type = validate_data_type(value_type)

    def _equals(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return DataType._equals(self, other) and \
            self.key_type == other.key_type and \
            self.value_type == other.value_type

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return isinstance(other, Dict) and \
            self.key_type == other.key_type and \
            self.value_type == other.value_type and \
            self.nullable == other.nullable


class Struct(_Struct):
    __slots__ = ()
    CLASS_NAME = "Struct"

    def __init__(self, field_types, nullable=True):
        DataType.__init__(self, nullable=nullable)
        self.field_types = OrderedDict()
        if isinstance(field_types, dict):
            field_types = six.iteritems(field_types)
        for k, v in field_types:
            self.field_types[k] = validate_data_type(v)
        self.namedtuple_type = xnamedtuple(
            "StructNamedTuple", list(self.field_types.keys())
        )
        self._struct_as_dict = options.struct_as_dict

    def _equals(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return isinstance(other, Struct) and \
            len(self.field_types) == len(other.field_types) and \
            all(self.field_types[k] == other.field_types[k] for k in six.iterkeys(self.field_types))

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return isinstance(other, Struct) and self == other and \
            self.nullable == other.nullable


int8 = Int8()
int16 = Int16()
int32 = Int32()
int64 = Int64()
float32 = Float32()
float64 = Float64()
boolean = Boolean()
string = String()
decimal = Decimal()
datetime = Datetime()
date = Date()
timestamp = Timestamp()
binary = Binary()
json = Json()


_data_types = dict(
    (t.name, t) for t in
    (int8, int16, int32, int64, float32, float64,
     boolean, string, decimal, datetime, binary,
     date, timestamp)
)


_composite_handlers = dict(
    list=List,
    dict=Dict,
    struct=Struct,
)


def validate_data_type(data_type):
    if isinstance(data_type, DataType):
        return data_type

    if isinstance(data_type, type):
        data_type = data_type.__name__

    if isinstance(data_type, six.string_types):
        data_type = data_type.lower()
        if data_type == 'int':
            data_type = 'int64'
        elif data_type == 'float':
            data_type = 'float64'
        elif data_type == 'bool':
            data_type = 'boolean'
        if data_type in _data_types:
            return _data_types[data_type]

        composite_type = parse_composite_types(data_type, _composite_handlers)
        if composite_type:
            return composite_type

    raise ValueError('Invalid data type: %s' % repr(data_type))


def validate_value_type(value, data_type=None):
    try:
        from pandas import Timestamp
    except ImportError:
        Timestamp = None

    if data_type is not None:
        data_type.validate_value(value)
        return data_type

    inferred_value_type = None

    if isinstance(value, bool):
        inferred_value_type = boolean
    elif isinstance(value, six.integer_types):
        for t in (int8, int16, int32, int64):
            if t.validate_value(value):
                inferred_value_type = t
                break
        if inferred_value_type is None:
            raise ValueError('Integer value too large: %s' % value)
    elif isinstance(value, float):
        inferred_value_type = float64
    elif isinstance(value, six.string_types):
        inferred_value_type = string
    elif isinstance(value, _Decimal):
        inferred_value_type = decimal
    elif isinstance(value, _datetime):
        inferred_value_type = datetime
    elif isinstance(value, _date):
        inferred_value_type = date
    elif Timestamp is not None and isinstance(value, Timestamp):
        inferred_value_type = timestamp
    else:
        raise ValueError('Unknown value: %s, type: %s' % (value, type(value)))

    return inferred_value_type


_number_types = OrderedDict.fromkeys([
    int8, int16, int32, int64, float32, float64, decimal])


def number_types():
    return _number_types.keys()


def is_number(data_type):
    if not isinstance(data_type, DataType):
        data_type = validate_data_type(data_type)

    if data_type in _number_types:
        return True

    return False


class Unknown(DataType):
    __slots__ = 'type', # the type of the column and identify that it's a dynamic field
    _singleton = False
    CLASS_NAME = 'Unknown'

    def __init__(self, nullable=True, type=None):
        super(Unknown, self).__init__(nullable=nullable)
        self.type = type

    def _equals(self, other):
        # ``Unknown`` type is not equal to other types
        return False

    def can_implicit_cast(self, other):
        # ``Unknown`` can cast to other types
        return True


class DynamicSchema(TableSchema):
    def __init__(self, *args, **kwargs):
        self.default_type = kwargs.pop('default_type', None)
        super(DynamicSchema, self).__init__(*args, **kwargs)

    def __contains__(self, item):
        # We do not know the actual columns,
        # just return True
        return True

    def __eq__(self, other):
        return False

    def __getitem__(self, item):
        if isinstance(item, six.string_types):
            try:
                return super(DynamicSchema, self).__getitem__(item)
            except ValueError:
                return Column(name=item, type=Unknown(type=self.default_type))

        return super(DynamicSchema, self).__getitem__(item)

    def get_column(self, name):
        try:
            return super(DynamicSchema, self).get_column()
        except ValueError:
            return Column(name=name, type=Unknown(type=self.default_type))

    @classmethod
    def from_schema(cls, schema, default_type=None):
        if isinstance(schema, DynamicSchema):
            if default_type == schema.default_type:
                return schema
            default_type = default_type or schema.default_type
            return DynamicSchema(columns=schema._columns,
                                 default_type=default_type)
        return DynamicSchema(columns=schema._columns,
                             default_type=default_type)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import base64
import sys
import json
import socket
import subprocess

from .client import GS_COORDINATOR_NAME


DEFAULT_GS_COORDINATOR_PORT = 63800
DEFAULT_GS_COORDINATOR_GATEWAY_PORT = 64800


def start_coordinator(args, coordinator_port, vineyard_socket):
    subprocess.Popen(
        [
            sys.executable,
            "-m",
            "gscoordinator",
            "--enable_k8s",
            "false",
            "--hosts",
            ",".join(args["worker_pod_ip_list"]),
            "--port",
            str(coordinator_port),
            "--num_workers",
            str(args["num_workers"]),
            "--vineyard_socket",
            vineyard_socket,
            "--log_level",
            "DEBUG",
            "--timeout_seconds",
            "-1",
        ],
        stdout=sys.__stdout__,
        stderr=sys.__stdout__,
        bufsize=1,
    )


def start_coordinator_gateway(args, coordinator_port, coordinator_gateway_port):
    subprocess.Popen(
        [
            "/usr/local/bin/graphscope-gateway",
            "-grpc-server-endpoint",
            "localhost:%s" % coordinator_port,
            "--gateway-address",
            ":%s" % coordinator_gateway_port,
            "-v",
            "100",
            "-alsologtostderr",
        ],
        stdout=sys.__stdout__,
        stderr=sys.__stdout__,
        bufsize=1,
    )


def _main():
    argv = sys.argv[1]
    args_dict = json.loads(base64.b64decode(argv).decode())
    print("launch graphscope:", args_dict)

    from cupid import context

    cupid_context = context()
    host_addr = socket.gethostbyname(socket.gethostname())

    os.environ.pop("KUBE_API_ADDRESS")

    coordinator_port = args_dict.get("port", None) or DEFAULT_GS_COORDINATOR_PORT
    coordinator_gateway_port = (
        args_dict.get("gateway_port", None) or DEFAULT_GS_COORDINATOR_GATEWAY_PORT
    )

    endpoint = "http://{0}:{1}".format(host_addr, coordinator_gateway_port)
    kvstore = cupid_context.kv_store()
    kvstore[GS_COORDINATOR_NAME] = json.dumps(dict(endpoint=endpoint))

    # start coordinator
    vineyard_socket = os.environ.get("VINEYARD_IPC_SOCKET", "/tmp/vineyard.sock")
    start_coordinator(args_dict, coordinator_port, vineyard_socket)
    start_coordinator_gateway(args_dict, coordinator_port, coordinator_gateway_port)

    # modify in hyper mode
    if os.environ.get("VM_ENGINE_TYPE") == "hyper":
        endpoint = socket.gethostname() + "-{}".format(coordinator_port)
    cupid_context.register_application(GS_COORDINATOR_NAME, endpoint)


if __name__ == "__main__":
    _main()
